#! /usr/bin/env python
"""Indexing and retrieving records locally."""
from __future__ import annotations

import collections
import hashlib
import json
import os
import sqlite3
import typing
from copy import deepcopy
from datetime import timedelta
from multiprocessing import Lock
from pathlib import Path
from threading import Timer

import git
import pandas as pd
import requests_cache
from git.exc import GitCommandError
from pybtex.database.input import bibtex
from thefuzz import fuzz
from tqdm import tqdm

import colrev.dataset
import colrev.env.environment_manager
import colrev.env.resources
import colrev.env.tei_parser
import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record
import colrev.ui_cli.cli_colors as colors


# import binascii

# pylint: disable=too-many-lines


class LocalIndex:
    """The LocalIndex implements indexing and retrieval of records across projects"""

    global_keys = ["doi", "colrev.dblp.dblp_key", "colrev_pdf_id", "url", "colrev_id"]
    max_len_sha256 = 2**256
    request_timeout = 90

    local_environment_path = Path.home().joinpath("colrev")
    SQLITE_PATH = str(local_environment_path / Path("sqlite_index.db"))

    teiind_path = local_environment_path / Path(".tei_index/")
    annotators_path = local_environment_path / Path("annotators")

    __sqlite_available = True

    sqlite_connection = None

    # Note : records are indexed by id = hash(colrev_id)
    # to ensure that the indexing-ids do not exceed limits
    # such as the opensearch limit of 512 bytes.
    # This enables efficient retrieval based on id=hash(colrev_id)
    # but also search-based retrieval using only colrev_ids

    RECORD_INDEX = "record_index"
    TOC_INDEX = "toc_index"
    UPDATE_LAYERD_FIELDS_QUERY = """
            UPDATE record_index SET
            layered_fields=?
            WHERE id=?"""

    SELECT_LAYERD_FIELDS_QUERY = "SELECT layered_fields FROM record_index WHERE id=?"

    SELECT_ALL_QUERIES = {
        TOC_INDEX: "SELECT * FROM toc_index WHERE",
        RECORD_INDEX: "SELECT * FROM record_index WHERE",
    }

    SELECT_KEY_QUERIES = {
        (RECORD_INDEX, "id"): "SELECT * FROM record_index WHERE id=?",
        (TOC_INDEX, "toc_key"): "SELECT * FROM toc_index WHERE toc_key=?",
        (RECORD_INDEX, "colrev_id"): "SELECT * FROM record_index WHERE colrev_id=?",
        (RECORD_INDEX, "doi"): "SELECT * FROM record_index where doi=?",
        (
            RECORD_INDEX,
            "colrev.dblp.dblp_key",
        ): "SELECT * FROM record_index WHERE dblp_key=?",
        (
            RECORD_INDEX,
            "colrev_pdf_id",
        ): "SELECT * FROM record_index WHERE colrev_pdf_id=?",
        (RECORD_INDEX, "url"): "SELECT * FROM record_index WHERE url=?",
    }

    # AUTHOR_INDEX = "author_index"
    # AUTHOR_RECORD_INDEX = "author_record_index"
    # CITATIONS_INDEX = "citations_index"

    RECORDS_INDEX_KEYS = [
        "id",
        "colrev_id",
        "citation_key",
        "title",
        "abstract",
        "file",
        "tei",
        "fulltext",
        "url",
        "doi",
        "dblp_key",  # Note : no dots in key names
        "colrev_pdf_id",
        "bibtex",
        "layered_fields"
        # "curation_ID"
    ]

    # Note: we need the local_curated_metadata field for is_duplicate()

    def __init__(
        self,
        *,
        index_tei: bool = False,
        verbose_mode: bool = False,
    ) -> None:
        self.verbose_mode = verbose_mode
        self.environment_manager = colrev.env.environment_manager.EnvironmentManager()
        self.__index_tei = index_tei

        self.thread_lock = Lock()

    def __get_sqlite_cursor(self, *, init: bool = False) -> sqlite3.Cursor:
        if init:
            Path(self.SQLITE_PATH).unlink(missing_ok=True)

        self.sqlite_connection = sqlite3.connect(self.SQLITE_PATH, timeout=90)
        self.sqlite_connection.row_factory = self.__dict_factory
        return self.sqlite_connection.cursor()

    def load_journal_rankings(self) -> None:
        """Loads journal rankings into sqlite database"""

        print("Index rankings")

        rankings_csv_path = (
            str(Path(__file__).parents[1])
            / Path("template")
            / Path("ops")
            / Path("journal_rankings.csv")
        )
        conn = sqlite3.connect(self.SQLITE_PATH)
        data_frame = pd.read_csv(rankings_csv_path, encoding="utf-8")
        data_frame.to_sql("rankings", conn, if_exists="replace", index=False)
        conn.commit()
        conn.close()

    def search_in_database(self, journal: typing.Optional[typing.Any]) -> list:
        """Searches for journalranking in database"""
        cur = self.__get_sqlite_cursor(init=False)
        cur.execute(
            "SELECT * FROM rankings WHERE journal_name = ?",
            (journal,),
        )
        rankings = cur.fetchall()
        return rankings

    def __dict_factory(self, cursor: sqlite3.Cursor, row: dict) -> dict:
        ret_dict = {}
        for idx, col in enumerate(cursor.description):
            ret_dict[col[0]] = row[idx]
        return ret_dict

    def __get_record_hash(self, *, record_dict: dict) -> str:
        # Note : may raise NotEnoughDataToIdentifyException
        string_to_hash = colrev.record.Record(data=record_dict).create_colrev_id()
        return hashlib.sha256(string_to_hash.encode("utf-8")).hexdigest()

    # def __increment_hash(self, *, paper_hash: str) -> str:
    #     plaintext = binascii.unhexlify(paper_hash)
    #     # also, we'll want to know our length later on
    #     plaintext_length = len(plaintext)
    #     plaintext_number = int.from_bytes(plaintext, "big")

    #     # recommendation: do not increment by 1
    #     plaintext_number += 10
    #     plaintext_number = plaintext_number % self.max_len_sha256

    #     new_plaintext = plaintext_number.to_bytes(plaintext_length, "big")
    #     new_hex = binascii.hexlify(new_plaintext)
    #     # print(new_hex.decode("utf-8"))

    #     return new_hex.decode("utf-8")

    def __get_tei_index_file(self, *, paper_hash: str) -> Path:
        return self.teiind_path / Path(f"{paper_hash[:2]}/{paper_hash[2:]}.tei.xml")

    # def __index_author(
    #     self, tei: colrev.env.tei_parser.TEIParser, record_dict: dict
    # ) -> None:
    #     print("index_author currently not implemented")
    # author_details = tei.get_author_details()
    # # Iterate over curated metadata and enrich it based on TEI (may vary in quality)
    # for author in record_dict.get("author", "").split(" and "):
    #     if "," not in author:
    #         continue
    #     author_dict = {}
    #     author_dict["surname"] = author.split(", ")[0]
    #     author_dict["forename"] = author.split(", ")[1]
    #     for author_detail in author_details:
    #         if author_dict["surname"] == author_detail["surname"]:
    #             # Add complementary details
    #             author_dict = {**author_dict, **author_detail}
    #     self.open_search.index(index=self.AUTHOR_INDEX, body=author_dict)

    def __index_tei_document(self, *, recs_to_index: list) -> None:
        if not self.__index_tei:
            return
        for record_dict in recs_to_index:
            if not Path(record_dict.get("file", "NA")).is_file():
                continue

            try:
                paper_hash = record_dict["id"]
                tei_path = self.__get_tei_index_file(paper_hash=paper_hash)
                tei_path.parents[0].mkdir(exist_ok=True, parents=True)
                if not tei_path.is_file():
                    print(f"Create tei for {record_dict['file']}")
                tei = colrev.env.tei_parser.TEIParser(
                    environment_manager=self.environment_manager,
                    pdf_path=Path(record_dict["file"]),
                    tei_path=tei_path,
                )

                record_dict["tei"] = str(tei_path)
                record_dict["fulltext"] = tei.get_tei_str()

                # self.__index_author(tei=tei, record_dict=record_dict)

            except (
                colrev_exceptions.TEIException,
                AttributeError,
                FileNotFoundError,
                colrev_exceptions.ServiceNotAvailableException,
            ):  # pragma: no cover
                pass

    def __amend_record(
        self, *, cur: sqlite3.Cursor, item: dict, curated_fields: list
    ) -> None:
        """Adds layered fields to amend existing records"""

        record_dict = self.__get_record_from_row(row=item)

        layered_fields = []
        cur.execute(self.SELECT_LAYERD_FIELDS_QUERY, (item["id"],))
        row = cur.fetchone()
        if row["layered_fields"]:
            layered_fields = json.loads(row["layered_fields"])
        for curated_field in curated_fields:
            if curated_field not in record_dict:
                continue
            source = record_dict["colrev_data_provenance"][curated_field]["source"]
            layered_fields.append(
                {
                    "key": curated_field,
                    "value": record_dict[curated_field],
                    "source": source,
                }
            )

        cur.execute(
            self.UPDATE_LAYERD_FIELDS_QUERY,
            (json.dumps(layered_fields), item["id"]),
        )

    def get_fields_to_remove(self, *, record_dict: dict) -> list:
        """Compares the record to available toc items and
        returns fields to remove (if any)"""
        # pylint: disable=too-many-return-statements

        fields_to_remove: typing.List[str] = []
        if "journal" not in record_dict and record_dict["ENTRYTYPE"] != "article":
            return fields_to_remove

        internal_record_dict = deepcopy(record_dict)

        if all(x in internal_record_dict.keys() for x in ["volume", "number"]):
            try:
                toc_key_full = colrev.record.Record(
                    data=internal_record_dict
                ).get_toc_key()

                if self.__toc_exists(toc_item=toc_key_full):
                    return fields_to_remove
            except colrev_exceptions.NotTOCIdentifiableException:
                return fields_to_remove
            wo_nr = deepcopy(internal_record_dict)
            del wo_nr["number"]
            toc_key_wo_nr = colrev.record.Record(data=wo_nr).get_toc_key()

            if toc_key_wo_nr != "NA":
                if self.__toc_exists(toc_item=toc_key_wo_nr):
                    fields_to_remove.append("number")
                    return fields_to_remove

            wo_vol = deepcopy(internal_record_dict)
            del wo_vol["volume"]
            toc_key_wo_vol = colrev.record.Record(data=wo_vol).get_toc_key()
            if toc_key_wo_vol != "NA":
                if self.__toc_exists(toc_item=toc_key_wo_vol):
                    fields_to_remove.append("volume")
                    return fields_to_remove

            wo_vol_nr = deepcopy(internal_record_dict)
            del wo_vol_nr["volume"]
            del wo_vol_nr["number"]
            toc_key_wo_vol_nr = colrev.record.Record(data=wo_vol_nr).get_toc_key()
            if toc_key_wo_vol_nr != "NA":
                if self.__toc_exists(toc_item=toc_key_wo_vol_nr):
                    fields_to_remove.append("number")
                    fields_to_remove.append("volume")
                    return fields_to_remove

        return fields_to_remove

    def __add_index_toc(self, *, toc_to_index: dict) -> None:
        list_to_add = list((k, v) for k, v in toc_to_index.items() if v != "DROPPED")
        if not self.sqlite_connection:
            return
        cur = self.sqlite_connection.cursor()
        try:
            cur.executemany(f"INSERT INTO {self.TOC_INDEX} VALUES(?, ?)", list_to_add)
        except sqlite3.IntegrityError as exc:
            if self.verbose_mode:
                print(exc)
        finally:
            if self.sqlite_connection:
                self.sqlite_connection.commit()

    def __add_index_records(self, *, recs_to_index: list, curated_fields: list) -> None:
        list_to_add = [
            {k: v for k, v in el.items() if k in self.RECORDS_INDEX_KEYS}
            for el in recs_to_index
        ]

        if not self.sqlite_connection:
            return
        cur = self.sqlite_connection.cursor()
        for item in list_to_add:
            while True:
                for records_index_required_key in self.RECORDS_INDEX_KEYS:
                    if records_index_required_key not in item:
                        item[records_index_required_key] = ""
                if item["id"] == "":
                    print("NO ID IN RECORD")
                    break
                try:
                    cur.execute(
                        f"INSERT INTO {self.RECORD_INDEX} "
                        f"VALUES(:{', :'.join(self.RECORDS_INDEX_KEYS)})",
                        item,
                    )
                    break
                except sqlite3.IntegrityError:
                    if not curated_fields:
                        break
                    try:
                        stored_record = self.__get_item_from_index(
                            index_name=self.RECORD_INDEX,
                            key="colrev_id",
                            value=item["colrev_id"],
                        )
                        stored_colrev_id = colrev.record.Record(
                            data=stored_record
                        ).create_colrev_id()

                        if stored_colrev_id == item["colrev_id"]:
                            self.__amend_record(
                                cur=cur, item=item, curated_fields=curated_fields
                            )
                            break

                        print("Collisions (TODO):")
                        print(stored_colrev_id)
                        print(item["colrev_id"])

                        # print(
                        #     [
                        #         {k: v for k, v in x.items() if k != "bibtex"}
                        #         for x in stored_record
                        #     ]
                        # )
                        # print(item)
                        # to handle the collision:
                        # print(f"Collision: {paper_hash}")
                        # print(cid_to_index)
                        # print(saved_record_cid)
                        # print(saved_record)
                        # paper_hash = self.__increment_hash(paper_hash=paper_hash)
                        # item["id"] = paper_hash
                        # continue in while-loop/try to insert...
                    except colrev_exceptions.RecordNotInIndexException:
                        break

        if self.sqlite_connection:
            self.sqlite_connection.commit()

    def __get_record_from_row(self, *, row: dict) -> dict:
        parser = bibtex.Parser()
        bib_data = parser.parse_string(row["bibtex"])
        ret = colrev.dataset.Dataset.parse_records_dict(records_dict=bib_data.entries)
        retrieved_record = list(ret.values())[0]

        # append layered fields
        if row["layered_fields"]:
            layered_fields = json.loads(row["layered_fields"])
            for layered_field in layered_fields:
                retrieved_record[layered_field["key"]] = layered_field["value"]
                if "colrev_data_provenance" not in retrieved_record:
                    retrieved_record["colrev_data_provenance"] = {}
                retrieved_record["colrev_data_provenance"][layered_field["key"]] = {
                    "source": layered_field["source"],
                    "note": "",
                }
        return retrieved_record

    def __retrieve_based_on_colrev_id(self, *, cids_to_retrieve: list) -> dict:
        # Note : may raise NotEnoughDataToIdentifyException

        for cid_to_retrieve in cids_to_retrieve:
            try:
                retrieved_record = self.__get_item_from_index(
                    index_name=self.RECORD_INDEX,
                    key="colrev_id",
                    value=cid_to_retrieve,
                )
                return retrieved_record

            except colrev_exceptions.RecordNotInIndexException:
                continue  # continue with the next cid_to_retrieve

        raise colrev_exceptions.RecordNotInIndexException()

    def _retrieve_from_github_curation(
        self, *, record_dict: dict
    ) -> dict:  # pragma: no cover
        ret = {}
        try:
            gh_url, record_id = record_dict["curation_ID"].split("#")

            temp_path = Path.home().joinpath("colrev").joinpath("test")
            temp_path.mkdir(exist_ok=True, parents=True)
            target_path = Path(temp_path) / Path(gh_url.split("/")[-1])
            if not target_path.is_dir():
                git.Repo.clone_from(
                    gh_url,  # .replace("https://github.com/", "git@github.com:") + ".git",
                    str(target_path),
                    depth=1,
                )

            content = ""
            with open(target_path / Path("data/records.bib"), encoding="utf-8") as file:
                content = file.read()

            parser = bibtex.Parser()
            bib_data = parser.parse_string(content)
            ret = colrev.dataset.Dataset.parse_records_dict(
                records_dict=bib_data.entries
            )
        except GitCommandError:
            pass

        if record_id not in ret:
            raise colrev_exceptions.RecordNotInIndexException

        ret[record_id]["curation_ID"] = record_dict["curation_ID"]
        return ret[record_id]

    def __retrieve_from_record_index(self, *, record_dict: dict) -> dict:
        # Note : may raise NotEnoughDataToIdentifyException

        record = colrev.record.Record(data=record_dict)

        if not self.__sqlite_available:
            if record_dict.get("curation_ID", "NA").startswith("https://github.com/"):
                return self._retrieve_from_github_curation(record_dict=record_dict)
            raise colrev_exceptions.RecordNotInIndexException

        if "colrev_id" in record.data:
            cid_to_retrieve = record.get_colrev_id()
        else:
            cid_to_retrieve = [record.create_colrev_id(assume_complete=True)]

        retrieved_record = self.__retrieve_based_on_colrev_id(
            cids_to_retrieve=cid_to_retrieve
        )
        if retrieved_record["ENTRYTYPE"] != record_dict["ENTRYTYPE"]:
            raise colrev_exceptions.RecordNotInIndexException
        return retrieved_record

    def __prepare_record_for_return(
        self,
        *,
        record_dict: dict,
        include_file: bool = False,
        include_colrev_ids: bool = False,
    ) -> dict:
        """Prepare a record for return (from local index)"""

        # pylint: disable=too-many-branches

        # Note : remove fulltext before parsing because it raises errors
        fulltext_backup = record_dict.get("fulltext", "NA")

        keys_to_remove = (
            "colrev_origin",
            "fulltext",
            "tei_file",
            "grobid-version",
            "excl_criteria",
            "exclusion_criteria",
            "screening_criteria",
            "local_curated_metadata",
            "metadata_source_repository_paths",
        )

        for key in keys_to_remove:
            record_dict.pop(key, None)

        # Note: record['file'] should be an absolute path by definition
        # when stored in the LocalIndex
        if "file" in record_dict and not Path(record_dict["file"]).is_file():
            del record_dict["file"]

        if not include_colrev_ids and "colrev_id" in record_dict:
            del record_dict["colrev_id"]

        if include_file:
            if fulltext_backup != "NA":
                record_dict["fulltext"] = fulltext_backup
        else:
            if "file" in record_dict:
                del record_dict["file"]
            if "file" in record_dict.get("colrev_data_provenance", {}):
                del record_dict["colrev_data_provenance"]["file"]
            if "colrev_pdf_id" in record_dict:
                del record_dict["colrev_pdf_id"]
            if "colrev_pdf_id" in record_dict.get("colrev_data_provenance", {}):
                del record_dict["colrev_data_provenance"]["colrev_pdf_id"]

        record = colrev.record.Record(data=record_dict)
        record.set_status(target_state=colrev.record.RecordState.md_prepared)

        if "CURATED" in record_dict.get("colrev_masterdata_provenance", {}):
            identifier_string = (
                record_dict["colrev_masterdata_provenance"]["CURATED"]["source"]
                + "#"
                + record_dict["ID"]
            )
            record_dict["curation_ID"] = identifier_string

        return record_dict

    def search(self, *, query: str) -> list[colrev.record.Record]:
        """Run a search for records"""

        records_to_return = []
        try:
            self.thread_lock.acquire(timeout=60)
            cur = self.__get_sqlite_cursor()
            selected_row = None
            print(f"{self.SELECT_ALL_QUERIES[self.RECORD_INDEX] } {query}")
            cur.execute(f"{self.SELECT_ALL_QUERIES[self.RECORD_INDEX] } {query}")
            for row in cur.fetchall():
                selected_row = row

                retrieved_record = self.__get_record_from_row(row=selected_row)

                retrieved_record = self.__prepare_record_for_return(
                    record_dict=retrieved_record, include_file=False
                )
                records_to_return.append(colrev.record.Record(data=retrieved_record))
        except sqlite3.OperationalError as exc:
            print(exc)
        finally:
            self.thread_lock.release()

        return records_to_return

    def __outlets_duplicated(self) -> bool:
        print("Validate curated metadata")

        try:
            curated_outlets = self.environment_manager.get_curated_outlets()

            if len(curated_outlets) != len(set(curated_outlets)):
                duplicated = [
                    item
                    for item, count in collections.Counter(curated_outlets).items()
                    if count > 1
                ]
                print(
                    f"Error: Duplicate outlets in curated_metadata : {','.join(duplicated)}"
                )
                return True

        except colrev_exceptions.CuratedOutletNotUnique as exc:
            print(exc)
            return True
        return False

    def __apply_status_requirements(self, *, record_dict: dict) -> None:
        if "colrev_status" not in record_dict:
            raise colrev_exceptions.RecordNotIndexableException()

        # It is important to exclude md_prepared if the LocalIndex
        # is used to dissociate duplicates
        if (
            record_dict["colrev_status"]
            in colrev.record.RecordState.get_non_processed_states()
        ):
            raise colrev_exceptions.RecordNotIndexableException()

        # Some prescreen_excluded records are not prepared
        if (
            record_dict["colrev_status"]
            == colrev.record.RecordState.rev_prescreen_excluded
        ):
            raise colrev_exceptions.RecordNotIndexableException()

    def __remove_fields(self, record_dict: dict) -> None:
        # Do not cover deprecated fields
        for deprecated_field in ["pdf_hash"]:
            if deprecated_field in record_dict:
                print(f"Removing deprecated field: {deprecated_field}")
                del record_dict[deprecated_field]

        if "screening_criteria" in record_dict:
            del record_dict["screening_criteria"]
        # Note: if the colrev_pdf_id has not been checked,
        # we cannot use it for retrieval or preparation.
        post_pdf_prepared_states = colrev.record.RecordState.get_post_x_states(
            state=colrev.record.RecordState.pdf_prepared
        )
        if record_dict["colrev_status"] not in post_pdf_prepared_states:
            if "colrev_pdf_id" in record_dict:
                del record_dict["colrev_pdf_id"]

        # Note : numbers of citations change regularly.
        # They should be retrieved from sources like crossref/doi.org
        if "cited_by" in record_dict:
            del record_dict["cited_by"]
        if record_dict.get("year", "NA").isdigit():
            record_dict["year"] = int(record_dict["year"])
        else:
            raise colrev_exceptions.RecordNotIndexableException()

        if "language" in record_dict and len(record_dict["language"]) != 3:
            print(f'Language not in ISO 639-3 format: {record_dict["language"]}')
            del record_dict["language"]

    def __adjust_provenance_for_indexint(self, *, record_dict: dict) -> None:
        # Provenance should point to the original repository path.
        # If the provenance/source was example.bib (and the record is amended during indexing)
        # we wouldn't know where the example.bib belongs to.
        record = colrev.record.Record(data=record_dict)
        for key in list(record.data.keys()):
            if (
                key
                not in colrev.record.Record.identifying_field_keys
                + colrev.record.Record.provenance_keys
                + [
                    "ID",
                    "ENTRYTYPE",
                    "local_curated_metadata",
                    "metadata_source_repository_paths",
                ]
            ):
                if key not in record.data.get("colrev_data_provenance", {}):
                    record.add_data_provenance(
                        key=key,
                        source=record_dict["metadata_source_repository_paths"],
                    )
                elif (
                    "CURATED"
                    not in record.data["colrev_data_provenance"][key]["source"]
                ):
                    record.add_data_provenance(
                        key=key,
                        source=record_dict["metadata_source_repository_paths"],
                    )
            elif not record.masterdata_is_curated():
                record.add_masterdata_provenance(
                    key=key, source=record_dict["metadata_source_repository_paths"]
                )

        # Make sure that we don't add provenance information without corresponding fields
        if "colrev_data_provenance" in record.data:
            provenance_keys = list(record.data.get("colrev_data_provenance", {}).keys())
            for provenance_key in provenance_keys:
                if provenance_key not in record.data:
                    del record.data["colrev_data_provenance"][provenance_key]
        if not record.masterdata_is_curated():
            if "colrev_masterdata_provenance" in record.data:
                provenance_keys = list(
                    record.data.get("colrev_masterdata_provenance", {}).keys()
                )
                for provenance_key in provenance_keys:
                    if provenance_key not in record.data:
                        del record.data["colrev_masterdata_provenance"][provenance_key]

        record_dict = record.get_data()

    def __prep_fields_for_indexing(self, *, record_dict: dict) -> None:
        # Note : this is the first run, no need to split/list
        if "colrev/curated_metadata" in record_dict["metadata_source_repository_paths"]:
            # Note : local_curated_metadata is important to identify non-duplicates
            # between curated_metadata_repositories
            record_dict["local_curated_metadata"] = "yes"

        # Note : file paths should be absolute when added to the LocalIndex
        if "file" in record_dict:
            pdf_path = Path(record_dict["file"])
            if pdf_path.is_file():
                record_dict["file"] = str(pdf_path)
            else:
                del record_dict["file"]

        if "colrev_origin" in record_dict:
            del record_dict["colrev_origin"]

        self.__adjust_provenance_for_indexint(record_dict=record_dict)

    def _prepare_record_for_indexing(self, *, record_dict: dict) -> dict:
        self.__apply_status_requirements(record_dict=record_dict)
        self.__remove_fields(record_dict=record_dict)
        self.__prep_fields_for_indexing(record_dict=record_dict)

        return record_dict

    def __get_index_record(self, *, record_dict: dict) -> dict:
        try:
            record_dict = self._prepare_record_for_indexing(record_dict=record_dict)
            cid_to_index = colrev.record.Record(data=record_dict).create_colrev_id()
            record_dict["colrev_id"] = cid_to_index
            record_dict["citation_key"] = record_dict["ID"]
            record_dict["id"] = self.__get_record_hash(record_dict=record_dict)
        except colrev_exceptions.NotEnoughDataToIdentifyException as exc:
            missing_key = ""
            if exc.missing_fields is not None:
                missing_key = ",".join(exc.missing_fields)
            raise colrev_exceptions.RecordNotIndexableException(
                missing_key=missing_key
            ) from exc

        return record_dict

    def __update_toc_index(
        self, *, toc_to_index: dict, copy_for_toc_index: dict, curated_masterdata: bool
    ) -> None:
        if not curated_masterdata or copy_for_toc_index.get("ENTRYTYPE", "") not in [
            "article",
            "inproceedings",
        ]:
            return

        toc_item = colrev.record.Record(data=copy_for_toc_index).get_toc_key()
        # Note : drop (do not index) tocs where records are missing
        # otherwise, record-not-in-toc will be triggered erroneously.
        drop_toc = copy_for_toc_index[
            "colrev_status"
        ] not in colrev.record.RecordState.get_post_x_states(
            state=colrev.record.RecordState.md_processed
        )
        try:
            colrev_id = colrev.record.Record(data=copy_for_toc_index).create_colrev_id(
                assume_complete=True
            )
        except colrev_exceptions.NotEnoughDataToIdentifyException:
            drop_toc = True
        if drop_toc:
            toc_to_index[toc_item] = "DROPPED"
        elif toc_to_index.get("toc_item", "") != "DROPPED":
            if toc_item in toc_to_index:
                toc_to_index[toc_item] += f";{colrev_id}"
            else:
                toc_to_index[toc_item] = colrev_id

    def index_records(
        self,
        *,
        records: dict,
        repo_source_path: Path,
        curation_url: str,
        curated_masterdata: bool,
        curated_fields: list,
    ) -> None:
        """Index a CoLRev project"""

        recs_to_index = []
        toc_to_index: typing.Dict[str, str] = {}
        for record_dict in tqdm(records.values()):
            copy_for_toc_index = deepcopy(record_dict)
            try:
                # Add metadata_source_repository_paths : list of repositories from which
                # the record was integrated. Important for is_duplicate(...)
                record_dict.update(
                    metadata_source_repository_paths=str(repo_source_path)
                )

                if curated_fields:
                    for curated_field in curated_fields:
                        colrev.record.Record(data=record_dict).add_data_provenance(
                            key=curated_field, source=f"CURATED:{curation_url}"
                        )
                if curated_masterdata:
                    record_dict.update(
                        colrev_masterdata_provenance=f"CURATED:{curation_url};;"
                    )

                # Set absolute file paths and set bibtex field (for simpler retrieval)
                if "file" in record_dict:
                    record_dict.update(
                        file=repo_source_path / Path(record_dict["file"])
                    )
                record_dict["bibtex"] = colrev.dataset.Dataset.parse_bibtex_str(
                    recs_dict_in={record_dict["ID"]: record_dict}
                )
                record_dict = self.__get_index_record(record_dict=record_dict)
                recs_to_index.append(record_dict)

            except (
                colrev_exceptions.RecordNotIndexableException,
                colrev_exceptions.NotTOCIdentifiableException,
                colrev_exceptions.NotEnoughDataToIdentifyException,
            ) as exc:
                if self.verbose_mode:
                    print(exc)
                    print(record_dict)
            finally:
                self.__update_toc_index(
                    toc_to_index=toc_to_index,
                    copy_for_toc_index=copy_for_toc_index,
                    curated_masterdata=curated_masterdata,
                )
        # Select fields and insert into index (sqlite)
        self.__index_tei_document(recs_to_index=recs_to_index)
        self.__add_index_records(
            recs_to_index=recs_to_index, curated_fields=curated_fields
        )
        if curated_masterdata:
            self.__add_index_toc(toc_to_index=toc_to_index)

    def __load_masterdata_curations(self) -> dict:  # pragma: no cover
        # Note : the following should be replaced by heuristics
        # based on the data (between colrev load and prep)
        masterdata_curations = {}
        filedata = colrev.env.utils.get_package_file_content(
            file_path=Path("template/ops/masterdata_curations.csv")
        )

        if filedata:
            for masterdata_curation in filedata.decode("utf-8").splitlines():
                masterdata_curations[
                    masterdata_curation.lower()
                ] = masterdata_curation.lower()

        return masterdata_curations

    def reinitialize_sqlite_db(self) -> None:
        """Reinitialize the SQLITE database ()"""
        print(f"Reinitialize {self.RECORD_INDEX} and {self.TOC_INDEX}")
        # Note : the tei-directory should be removed manually.

        cur = self.__get_sqlite_cursor(init=True)
        cur.execute(f"drop table if exists {self.RECORD_INDEX}")
        cur.execute(
            f"CREATE TABLE {self.RECORD_INDEX}(id TEXT PRIMARY KEY, "
            + ",".join(self.RECORDS_INDEX_KEYS[1:])
            + ")"
        )
        cur.execute(f"drop table if exists {self.TOC_INDEX}")
        cur.execute(
            f"CREATE TABLE {self.TOC_INDEX}(toc_key TEXT PRIMARY KEY, colrev_ids)"
        )
        if self.sqlite_connection:
            self.sqlite_connection.commit()

    def index_colrev_project(
        self, *, repo_source_path: Path
    ) -> None:  # pragma: no cover
        """Index a CoLRev project"""
        try:
            if not Path(repo_source_path).is_dir():
                print(f"Warning {repo_source_path} not a directory")
                return

            print(f"Index records from {repo_source_path}")
            os.chdir(repo_source_path)
            review_manager = colrev.review_manager.ReviewManager(
                path_str=str(repo_source_path)
            )

            check_operation = colrev.operation.CheckOperation(
                review_manager=review_manager
            )

            if review_manager.dataset.get_repo().active_branch.name != "main":
                print(
                    f"{colors.ORANGE}Warning: {repo_source_path} not on main branch{colors.END}"
                )

            if not check_operation.review_manager.dataset.records_file.is_file():
                return
            records = check_operation.review_manager.dataset.load_records_dict()

            curation_endpoints = [
                x
                for x in check_operation.review_manager.settings.data.data_package_endpoints
                if x["endpoint"] == "colrev.colrev_curation"
            ]

            curated_fields = []
            curation_url = ""
            if curation_endpoints:
                curation_endpoint = curation_endpoints[0]
                # Set masterdata_provenace to CURATED:{url}
                curation_url = curation_endpoint["curation_url"]
                if (
                    not check_operation.review_manager.settings.is_curated_masterdata_repo()
                ):
                    # Curated fields/layered_fields only for non-masterdata-curated repos
                    # Add curation_url to curated fields (provenance)
                    curated_fields = curation_endpoint["curated_fields"]

            curated_masterdata = (
                check_operation.review_manager.settings.is_curated_masterdata_repo()
            )

            self.index_records(
                records=records,
                repo_source_path=repo_source_path,
                curated_fields=curated_fields,
                curation_url=curation_url,
                curated_masterdata=curated_masterdata,
            )

        except colrev_exceptions.CoLRevException as exc:
            print(exc)

    def index(self) -> None:  # pragma: no cover
        """Index all registered CoLRev projects"""

        # pylint: disable=import-outside-toplevel
        # pylint: disable=redefined-outer-name
        # pylint: disable=cyclic-import
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-locals
        import colrev.review_manager

        # Note : this task takes long and does not need to run often
        session = requests_cache.CachedSession(
            str(colrev.env.environment_manager.EnvironmentManager.cache_path),
            backend="sqlite",
            expire_after=timedelta(days=30),
        )
        # Note : lambda is necessary to prevent immediate function call
        # pylint: disable=unnecessary-lambda
        Timer(0.1, lambda: session.remove_expired_responses()).start()

        if self.__outlets_duplicated():
            return

        self.reinitialize_sqlite_db()

        repo_source_paths = [
            x["repo_source_path"] for x in self.environment_manager.local_repos()
        ]
        if not repo_source_paths:
            env_resources = colrev.env.resources.Resources()
            curated_resources = list(self.__load_masterdata_curations().values())
            for curated_resource in curated_resources:
                print(f"Install {curated_resource}")
                env_resources.install_curated_resource(
                    curated_resource=curated_resource
                )

            repo_source_paths = [
                x["repo_source_path"] for x in self.environment_manager.local_repos()
            ]

        for repo_source_path in repo_source_paths:
            self.index_colrev_project(repo_source_path=repo_source_path)

        # for annotator in self.annotators_path.glob("*/annotate.py"):
        #     print(f"Load {annotator}")

        #     annotator_module = ....load_source("annotator_module", str(annotator))
        #     annotate = getattr(annotator_module, "annotate")
        #     annotate(self)
        # Note : es.update can use functions applied to each record (for the update)

    def get_year_from_toc(self, *, record_dict: dict) -> str:
        """Determine the year of a paper based on its table-of-content (journal-volume-number)"""

        try:
            toc_key = colrev.record.Record(data=record_dict).get_toc_key()
            toc_items = []
            if self.__toc_exists(toc_item=toc_key):
                res = self.__get_item_from_index(
                    index_name=self.TOC_INDEX, key="toc_key", value=toc_key
                )
                toc_items = res.get("colrev_ids", "").split(";")  # type: ignore

            if not toc_items:
                raise colrev_exceptions.TOCNotAvailableException()

            toc_records_colrev_id = toc_items[0]

            record_dict = self.__get_item_from_index(
                index_name=self.RECORD_INDEX,
                key="colrev_id",
                value=toc_records_colrev_id,
            )

            year = record_dict.get("year", "NA")

            return year

        except (
            colrev_exceptions.NotEnoughDataToIdentifyException,
            colrev_exceptions.NotTOCIdentifiableException,
            colrev_exceptions.RecordNotInIndexException,
        ) as exc:
            raise colrev_exceptions.TOCNotAvailableException() from exc

    def __toc_exists(self, *, toc_item: str) -> bool:
        try:
            self.thread_lock.acquire(timeout=60)
            cur = self.__get_sqlite_cursor()
            cur.execute(
                self.SELECT_KEY_QUERIES[(self.TOC_INDEX, "toc_key")], (toc_item,)
            )
            selected_row = cur.fetchone()
            self.thread_lock.release()
            if not selected_row:
                return False
            return True
        except sqlite3.OperationalError:
            self.thread_lock.release()
        except AttributeError:  # ie. no sqlite database available
            return False
        return False

    def __get_toc_items_for_toc_retrieval(
        self, *, toc_key: str, search_across_tocs: bool
    ) -> list:
        toc_items = []

        if self.__toc_exists(toc_item=toc_key):
            res = self.__get_item_from_index(
                index_name=self.TOC_INDEX, key="toc_key", value=toc_key
            )
            toc_items = res.get("colrev_ids", "").split(";")  # type: ignore
        else:
            if not search_across_tocs:
                raise colrev_exceptions.RecordNotInIndexException()

        if not toc_items and search_across_tocs:
            try:
                partial_toc_key = toc_key.replace("|-", "")
                retrieved_tocs = self.__get_items_from_index(
                    index_name=self.TOC_INDEX,
                    query=("toc_key LIKE ?", [f"{partial_toc_key}%"]),
                )
                toc_items = [x["colrev_ids"].split(";") for x in retrieved_tocs]
                toc_items = [item for sublist in toc_items for item in sublist]

            except (
                colrev_exceptions.NotTOCIdentifiableException,
                KeyError,
            ) as exc:
                raise colrev_exceptions.RecordNotInIndexException() from exc

        if not toc_items:
            raise colrev_exceptions.RecordNotInIndexException()
        return toc_items

    def retrieve_from_toc(
        self,
        *,
        record_dict: dict,
        similarity_threshold: float,
        include_file: bool = False,
        search_across_tocs: bool = False,
    ) -> dict:
        """Retrieve a record from the toc (table-of-contents)"""

        try:
            # Note: in NotTOCIdentifiableException cases, we still need a toc_key.
            # to accomplish this, the get_toc_key() may acced an "accept_incomplete" flag
            # try:
            toc_key = colrev.record.Record(data=record_dict).get_toc_key()
            # except colrev_exceptions.NotTOCIdentifiableException as exc:
            #     if not search_across_tocs:
            #         raise colrev_exceptions.RecordNotInIndexException() from exc

            toc_items = self.__get_toc_items_for_toc_retrieval(
                toc_key=toc_key, search_across_tocs=search_across_tocs
            )

            if search_across_tocs:
                record_colrev_id = colrev.record.Record(
                    data=record_dict
                ).create_colrev_id(assume_complete=True)

            else:
                record_colrev_id = colrev.record.Record(
                    data=record_dict
                ).create_colrev_id()

            sim_list = []

            for toc_records_colrev_id in toc_items:
                # Note : using a simpler similarity measure
                # because the publication outlet parameters are already identical
                sim_value = fuzz.ratio(record_colrev_id, toc_records_colrev_id) / 100
                sim_list.append(sim_value)

            if not sim_list:
                raise colrev_exceptions.RecordNotInTOCException(
                    record_id=record_dict["ID"], toc_key=toc_key
                )

            if max(sim_list) < similarity_threshold:
                raise colrev_exceptions.RecordNotInTOCException(
                    record_id=record_dict["ID"], toc_key=toc_key
                )

            if search_across_tocs:
                second_highest = list(set(sim_list))[-2]
                # Require a minimum difference to the next most similar record
                if (max(sim_list) - second_highest) < 0.2:
                    raise colrev_exceptions.RecordNotInIndexException()

            toc_records_colrev_id = toc_items[sim_list.index(max(sim_list))]

            record_dict = self.__get_item_from_index(
                index_name=self.RECORD_INDEX,
                key="colrev_id",
                value=toc_records_colrev_id,
            )

            return self.__prepare_record_for_return(
                record_dict=record_dict, include_file=include_file
            )

        except (
            colrev_exceptions.NotEnoughDataToIdentifyException,
            colrev_exceptions.NotTOCIdentifiableException,
        ):
            pass

        raise colrev_exceptions.RecordNotInIndexException()

    def __get_items_from_index(
        self, *, index_name: str, query: typing.Tuple[str, list[str]]
    ) -> list:
        try:
            self.thread_lock.acquire(timeout=60)
            cur = self.__get_sqlite_cursor()
            select_all_query = f"{self.SELECT_ALL_QUERIES[index_name]} {query[0]}"
            cur.execute(select_all_query, query[1])
            results = cur.fetchall()
            self.thread_lock.release()
            return results

        except sqlite3.OperationalError as exc:
            self.thread_lock.release()
            raise colrev_exceptions.RecordNotInIndexException() from exc
        except AttributeError as exc:
            raise colrev_exceptions.RecordNotInIndexException() from exc

    def __get_item_from_index(self, *, index_name: str, key: str, value: str) -> dict:
        try:
            self.thread_lock.acquire(timeout=60)
            cur = self.__get_sqlite_cursor()

            # in the following, collisions should be handled.
            # paper_hash = hashlib.sha256(cid_to_retrieve.encode("utf-8")).hexdigest()
            # Collision
            # paper_hash = self.__increment_hash(paper_hash=paper_hash)

            cur.execute(self.SELECT_KEY_QUERIES[(index_name, key)], (value,))

            selected_row = cur.fetchone()
            self.thread_lock.release()

            if not selected_row:
                raise colrev_exceptions.RecordNotInIndexException()

            retrieved_record = {}
            if self.RECORD_INDEX == index_name:
                retrieved_record = self.__get_record_from_row(row=selected_row)
            else:
                retrieved_record = selected_row

            if key == "colrev_id":
                if (
                    value
                    != colrev.record.Record(data=retrieved_record).create_colrev_id()
                ):
                    raise colrev_exceptions.RecordNotInIndexException()
            else:
                if key not in retrieved_record:
                    raise colrev_exceptions.RecordNotInIndexException()

                if value != retrieved_record[key]:
                    raise colrev_exceptions.RecordNotInIndexException()
            return retrieved_record

        except sqlite3.OperationalError as exc:
            self.thread_lock.release()
            raise colrev_exceptions.RecordNotInIndexException() from exc

    def retrieve_based_on_colrev_pdf_id(self, *, colrev_pdf_id: str) -> dict:
        """
        Convenience function to retrieve the indexed record_dict metadata
        based on a colrev_pdf_id
        """

        record_dict = self.__get_item_from_index(
            index_name=self.RECORD_INDEX,
            key="colrev_pdf_id",
            value=colrev_pdf_id,
        )

        record_to_import = self.__prepare_record_for_return(
            record_dict=record_dict, include_file=True
        )
        if "file" in record_to_import:
            del record_to_import["file"]
        return record_to_import

    def retrieve(
        self,
        *,
        record_dict: dict,
        include_file: bool = False,
        include_colrev_ids: bool = False,
    ) -> dict:
        """
        Convenience function to retrieve the indexed record_dict metadata
        based on another record_dict
        """

        # To avoid modifications to the original record
        record_dict = deepcopy(record_dict)

        retrieved_record_dict: typing.Dict = {}
        # 1. Try the record index
        try:
            retrieved_record_dict = self.__retrieve_from_record_index(
                record_dict=record_dict
            )
        except (
            colrev_exceptions.RecordNotInIndexException,
            colrev_exceptions.NotEnoughDataToIdentifyException,
        ) as exc:
            if self.verbose_mode:
                print(exc)
                print(f"{record_dict['ID']} - no exact match")

        # 2. Try using global-ids
        if not retrieved_record_dict and self.__sqlite_available:
            remove_colrev_id = False
            if "colrev_id" not in record_dict:
                try:
                    record_dict["colrev_id"] = colrev.record.Record(
                        data=record_dict
                    ).create_colrev_id()
                    remove_colrev_id = True
                except colrev_exceptions.NotEnoughDataToIdentifyException:
                    pass
            for key, value in record_dict.items():
                if key not in self.global_keys or "ID" == key:
                    continue

                retrieved_record_dict = self.__get_item_from_index(
                    index_name=self.RECORD_INDEX, key=key, value=value
                )
                if key in retrieved_record_dict:
                    if retrieved_record_dict[key] == value:
                        break
                retrieved_record_dict = {}
            if remove_colrev_id:
                del record_dict["colrev_id"]

        if not retrieved_record_dict:
            raise colrev_exceptions.RecordNotInIndexException(
                record_dict.get("ID", "no-key")
            )

        return self.__prepare_record_for_return(
            record_dict=retrieved_record_dict,
            include_file=include_file,
            include_colrev_ids=include_colrev_ids,
        )

    def is_duplicate(self, *, record1_colrev_id: list, record2_colrev_id: list) -> str:
        """Convenience function to check whether two records are a duplicate"""

        try:
            # Ensure that we receive actual lists
            # otherwise, __retrieve_based_on_colrev_id iterates over a string and
            # open_search_thread_instance.search returns random results
            assert isinstance(record1_colrev_id, list)
            assert isinstance(record2_colrev_id, list)

            # Prevent errors caused by short colrev_ids/empty lists
            if (
                any(len(cid) < 20 for cid in record1_colrev_id)
                or any(len(cid) < 20 for cid in record2_colrev_id)
                or 0 == len(record1_colrev_id)
                or 0 == len(record2_colrev_id)
            ):
                return "unknown"

            # Easy case: the initial colrev_ids overlap => duplicate
            initial_colrev_ids_overlap = not set(record1_colrev_id).isdisjoint(
                list(record2_colrev_id)
            )
            if initial_colrev_ids_overlap:
                return "yes"

            # Retrieve records from LocalIndex and use that information
            # to decide whether the records are duplicates

            r1_index = self.__retrieve_based_on_colrev_id(
                cids_to_retrieve=record1_colrev_id
            )
            r2_index = self.__retrieve_based_on_colrev_id(
                cids_to_retrieve=record2_colrev_id
            )
            # Each record may originate from multiple repositories simultaneously
            # see integration of records in __amend_record(...)
            # This information is stored in metadata_source_repository_paths (list)

            r1_metadata_source_repository_paths = r1_index[
                "metadata_source_repository_paths"
            ].split("\n")
            r2_metadata_source_repository_paths = r2_index[
                "metadata_source_repository_paths"
            ].split("\n")

            # There are no duplicates within repositories
            # because we only index records that are md_processed or beyond
            # see conditions of index_record(...)

            # The condition that two records are in the same repository is True if
            # their metadata_source_repository_paths overlap.
            # This does not change if records are also in non-overlapping repositories

            same_repository = not set(r1_metadata_source_repository_paths).isdisjoint(
                set(r2_metadata_source_repository_paths)
            )

            # colrev_ids must be used instead of IDs
            # because IDs of original repositories
            # are not available in the integrated record

            colrev_ids_overlap = not set(
                colrev.record.Record(data=r1_index).get_colrev_id()
            ).isdisjoint(
                list(list(colrev.record.Record(data=r2_index).get_colrev_id()))
            )

            if same_repository:
                if colrev_ids_overlap:
                    return "yes"
                return "no"

            # Curated metadata repositories do not curate outlets redundantly,
            # i.e., there are no duplicates between curated repositories.
            # see __outlets_duplicated(...)

            different_curated_repositories = (
                "CURATED" in r1_index.get("colrev_masterdata_provenance", "")
                and "CURATED" in r2_index.get("colrev_masterdata_provenance", "")
                and (
                    r1_index.get("colrev_masterdata_provenance", "a")
                    != r2_index.get("colrev_masterdata_provenance", "b")
                )
            )

            if different_curated_repositories:
                return "no"

        except (
            colrev_exceptions.RecordNotInIndexException,
            colrev_exceptions.NotEnoughDataToIdentifyException,
        ):
            pass

        return "unknown"



#!/usr/bin/env python3
"""Settings of the CoLRev project."""
from __future__ import annotations

import dataclasses
import json
import typing
import warnings
from dataclasses import asdict
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING

import dacite
from dacite import from_dict
from dacite.exceptions import MissingValueError
from dacite.exceptions import WrongTypeError
from dataclasses_jsonschema import FieldEncoder
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.utils
import colrev.exceptions as colrev_exceptions

if TYPE_CHECKING:
    import colrev.review_manager
    import colrev.ops.search_feed


# Note : to avoid performance issues on startup (ReviewManager, parsing settings)
# the settings dataclasses should be in one file (13s compared to 0.3s)


# Project


class IDPattern(Enum):
    """The pattern for generating record IDs"""

    # pylint: disable=invalid-name
    first_author_year = "first_author_year"
    three_authors_year = "three_authors_year"

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_

    @classmethod
    def get_field_details(cls) -> typing.Dict:
        """Get the field details"""
        # pylint: disable=no-member
        return {"options": cls._member_names_, "type": "selection"}


@dataclass
class Author(JsonSchemaMixin):
    """Author of the review"""

    # pylint: disable=too-many-instance-attributes

    name: str
    initials: str
    email: str
    orcid: typing.Optional[str] = None
    contributions: typing.List[str] = dataclasses.field(default_factory=list)
    affiliations: typing.Optional[str] = None
    funding: typing.List[str] = dataclasses.field(default_factory=list)
    identifiers: typing.List[str] = dataclasses.field(default_factory=list)


@dataclass
class Protocol(JsonSchemaMixin):
    """Review protocol"""

    url: str


class ShareStatReq(Enum):
    """Record status requirements for sharing"""

    # pylint: disable=invalid-name
    none = "none"
    processed = "processed"
    screened = "screened"
    completed = "completed"

    @classmethod
    def get_field_details(cls) -> typing.Dict:
        """Get the field details"""
        # pylint: disable=no-member
        return {"options": cls._member_names_, "type": "selection"}

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_


@dataclass
class ProjectSettings(JsonSchemaMixin):
    """Project settings"""

    # pylint: disable=too-many-instance-attributes

    title: str
    __doc_title__ = "The title of the review"
    authors: typing.List[Author]
    keywords: typing.List[str]
    # status ? (development/published?)
    protocol: typing.Optional[Protocol]
    # publication: ... (reference, link, ....)
    review_type: str
    id_pattern: IDPattern
    share_stat_req: ShareStatReq
    delay_automated_processing: bool
    colrev_version: str
    auto_upgrade: bool

    def __str__(self) -> str:
        project_str = f"Review ({self.review_type}):"
        if self.title:
            project_str += f" {self.title}"
        return project_str


# Search


class SearchType(Enum):
    """Type of search source"""

    DB = "DB"
    TOC = "TOC"
    BACKWARD_SEARCH = "BACKWARD_SEARCH"
    FORWARD_SEARCH = "FORWARD_SEARCH"
    PDFS = "PDFS"
    OTHER = "OTHER"

    @classmethod
    def get_field_details(cls) -> typing.Dict:
        """Get the field details"""
        # pylint: disable=no-member
        return {"options": cls._member_names_, "type": "selection"}

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_

    def __str__(self) -> str:
        return f"{self.name}"


@dataclass
class SearchSource(JsonSchemaMixin):
    """Search source settings"""

    # pylint: disable=too-many-instance-attributes
    endpoint: str
    filename: Path
    search_type: SearchType
    search_parameters: dict
    comment: typing.Optional[str]

    def __init__(
        self,
        *,
        endpoint: str,
        filename: Path,
        search_type: SearchType,
        search_parameters: dict,
        comment: typing.Optional[str],
    ) -> None:
        self.endpoint = endpoint
        assert filename.parent.name == "search"
        assert filename.parent.parent.name == "data"
        self.filename = filename
        self.search_type = search_type
        self.search_parameters = search_parameters
        self.comment = comment

    def setup_for_load(
        self,
        *,
        source_records_list: typing.List[typing.Dict],
        imported_origins: typing.List[str],
    ) -> None:
        """Set the SearchSource up for the load process (initialize statistics)"""
        # pylint: disable=attribute-defined-outside-init
        # Note : define outside init because the following
        # attributes are temporary. They should not be
        # saved to settings.json.

        self.to_import = len(source_records_list)
        self.imported_origins: typing.List[str] = imported_origins
        self.len_before = len(imported_origins)
        self.source_records_list: typing.List[typing.Dict] = source_records_list

    def get_origin_prefix(self) -> str:
        """Get the corresponding origin prefix"""
        assert not any(x in str(self.filename.name) for x in [";", "/"])
        return (
            str(self.filename.name)
            .replace(str(colrev.review_manager.ReviewManager.SEARCHDIR_RELATIVE), "")
            .lstrip("/")
        )

    def is_md_source(self) -> bool:
        """Check whether the source is a metadata source (for preparation)"""

        return str(self.filename.name).startswith("md_")

    def is_quasi_md_source(self) -> bool:
        """Check whether the source is a quasi-metadata source
        (when search files are added without search_parameters/queries)"""

        return not bool(self.search_parameters)

    def get_dict(self) -> dict:
        """Get the dict of SearchSources (for endpoint initalization)"""

        exported_dict = asdict(
            self, dict_factory=colrev.env.utils.custom_asdict_factory
        )

        exported_dict["search_type"] = colrev.settings.SearchType[
            exported_dict["search_type"]
        ]
        exported_dict["filename"] = Path(exported_dict["filename"])

        return exported_dict

    def get_feed(
        self,
        review_manager: colrev.review_manager.ReviewManager,
        source_identifier: str,
        update_only: bool,
    ) -> colrev.ops.search_feed.GeneralOriginFeed:
        """Get a feed to add and update records"""
        # pylint: disable=import-outside-toplevel
        # pylint: disable=cyclic-import
        import colrev.ops.search_feed

        return colrev.ops.search_feed.GeneralOriginFeed(
            review_manager=review_manager,
            search_source=self,
            source_identifier=source_identifier,
            update_only=update_only,
        )

    def __str__(self) -> str:
        optional_comment = ""
        if self.comment:
            optional_comment = f"\n   comment:             {self.comment}"
        return (
            f"{self.endpoint} (type: {self.search_type}, "
            + f"filename: {self.filename})\n"
            + f"   search parameters:   {self.search_parameters}"
            + optional_comment
        )


@dataclass
class SearchSettings(JsonSchemaMixin):
    """Search settings"""

    retrieve_forthcoming: bool

    def __str__(self) -> str:
        return f" - retrieve_forthcoming: {self.retrieve_forthcoming}"


# Load


@dataclass
class LoadSettings(JsonSchemaMixin):
    """Load settings"""

    def __str__(self) -> str:
        return " - TODO"


# Prep


@dataclass
class PrepRound(JsonSchemaMixin):
    """Prep round settings"""

    name: str
    similarity: float
    prep_package_endpoints: list

    def __str__(self) -> str:
        short_list = [script["endpoint"] for script in self.prep_package_endpoints][:3]
        if len(self.prep_package_endpoints) > 3:
            short_list.append("...")
        return f"{self.name} (" + ",".join(short_list) + ")"


@dataclass
class PrepSettings(JsonSchemaMixin):
    """Prep settings"""

    fields_to_keep: typing.List[str]
    prep_rounds: typing.List[PrepRound]

    prep_man_package_endpoints: list

    defects_to_ignore: list

    def __str__(self) -> str:
        return (
            " - prep_rounds:\n   - "
            + "\n   - ".join([str(prep_round) for prep_round in self.prep_rounds])
            + f"\n - fields_to_keep: {self.fields_to_keep}"
        )


# Dedupe


class SameSourceMergePolicy(Enum):
    """Policy for applying merges within the same search source"""

    # pylint: disable=invalid-name
    prevent = "prevent"
    warn = "warn"
    apply = "apply"

    @classmethod
    def get_field_details(cls) -> typing.Dict:
        """Get the field details"""
        # pylint: disable=no-member
        return {"options": cls._member_names_, "type": "selection"}

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_


@dataclass
class DedupeSettings(JsonSchemaMixin):
    """Dedupe settings"""

    same_source_merges: SameSourceMergePolicy
    dedupe_package_endpoints: list

    def __str__(self) -> str:
        return (
            f" - same_source_merges: {self.same_source_merges}\n"
            + " - "
            + ",".join([s["endpoint"] for s in self.dedupe_package_endpoints])
        )


# Prescreen


@dataclass
class PrescreenSettings(JsonSchemaMixin):
    """Prescreen settings"""

    explanation: str
    prescreen_package_endpoints: list

    def __str__(self) -> str:
        return "- " + ",".join(
            [s["endpoint"] for s in self.prescreen_package_endpoints]
        )


# PDF get


class PDFPathType(Enum):
    """Policy for handling PDFs (create symlinks or copy files)"""

    # pylint: disable=invalid-name
    symlink = "symlink"
    copy = "copy"

    @classmethod
    def get_field_details(cls) -> typing.Dict:
        """Get the field details"""
        # pylint: disable=no-member
        return {"options": cls._member_names_, "type": "selection"}

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_


@dataclass
class PDFGetSettings(JsonSchemaMixin):
    """PDF get settings"""

    pdf_path_type: PDFPathType
    pdf_required_for_screen_and_synthesis: bool
    """With the pdf_required_for_screen_and_synthesis flag, the PDF retrieval
    can be specified as mandatory (true) or optional (false) for the following steps"""
    rename_pdfs: bool
    pdf_get_package_endpoints: list

    pdf_get_man_package_endpoints: list

    def __str__(self) -> str:
        return (
            f" - pdf_path_type: {self.pdf_path_type}"
            + "\n - "
            + ",".join([s["endpoint"] for s in self.pdf_get_package_endpoints])
        )


# PDF prep


@dataclass
class PDFPrepSettings(JsonSchemaMixin):
    """PDF prep settings"""

    keep_backup_of_pdfs: bool

    pdf_prep_package_endpoints: list

    pdf_prep_man_package_endpoints: list

    def __str__(self) -> str:
        return "- " + "\n- ".join(
            [s["endpoint"] for s in self.pdf_prep_package_endpoints]
        )


# Screen


class ScreenCriterionType(Enum):
    """Type of screening criterion"""

    # pylint: disable=invalid-name
    inclusion_criterion = "inclusion_criterion"
    exclusion_criterion = "exclusion_criterion"

    @classmethod
    def get_field_details(cls) -> typing.Dict:
        """Get the field details"""
        # pylint: disable=no-member
        return {"options": cls._member_names_, "type": "selection"}

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_

    def __str__(self) -> str:
        return self.name


@dataclass
class ScreenCriterion(JsonSchemaMixin):
    """Screen criterion"""

    explanation: str
    comment: typing.Optional[str]
    criterion_type: ScreenCriterionType

    def __str__(self) -> str:
        return f"{self.criterion_type} {self.explanation} ({self.explanation})"


@dataclass
class ScreenSettings(JsonSchemaMixin):
    """Screen settings"""

    explanation: typing.Optional[str]
    criteria: typing.Dict[str, ScreenCriterion]
    screen_package_endpoints: list

    def __str__(self) -> str:
        if self.criteria:
            return " - " + "\n - ".join([str(c) for c in self.criteria])
        return " -"


# Data


@dataclass
class DataSettings(JsonSchemaMixin):
    """Data settings"""

    data_package_endpoints: list

    def __str__(self) -> str:
        return " - " + "\n- ".join([s["endpoint"] for s in self.data_package_endpoints])


@dataclass
class Settings(JsonSchemaMixin):
    """CoLRev project settings"""

    # pylint: disable=too-many-instance-attributes

    project: ProjectSettings
    sources: typing.List[SearchSource]
    search: SearchSettings
    load: LoadSettings
    prep: PrepSettings
    dedupe: DedupeSettings
    prescreen: PrescreenSettings
    pdf_get: PDFGetSettings
    pdf_prep: PDFPrepSettings
    screen: ScreenSettings
    data: DataSettings

    def is_curated_repo(self) -> bool:
        """Check whether data is curated in this repository"""

        curation_endpoints = [
            x
            for x in self.data.data_package_endpoints
            if x["endpoint"] == "colrev.colrev_curation"
        ]
        return bool(curation_endpoints)

    def is_curated_masterdata_repo(self) -> bool:
        """Check whether the masterdata is curated in this repository"""

        curation_endpoints = [
            x
            for x in self.data.data_package_endpoints
            if x["endpoint"] == "colrev.colrev_curation"
        ]
        if curation_endpoints:
            curation_endpoint = curation_endpoints[0]
            if curation_endpoint["curated_masterdata"]:
                return True
        return False

    def __str__(self) -> str:
        return (
            str(self.project)
            + "\nSearch\n"
            + str(self.search)
            + "\nSources\n- "
            + "\n- ".join([str(s) for s in self.sources])
            + "\nLoad\n"
            + str(self.load)
            + "\nPreparation\n"
            + str(self.prep)
            + "\nDedupe\n"
            + str(self.dedupe)
            + "\nPrescreen\n"
            + str(self.prescreen)
            + "\nPDF get\n"
            + str(self.pdf_get)
            + "\nPDF prep\n"
            + str(self.pdf_prep)
            + "\nScreen\n"
            + str(self.screen)
            + "\nData\n"
            + str(self.data)
        )

    @classmethod
    def get_settings_schema(cls) -> dict:
        """Get the json-schema for the settings"""

        class PathField(FieldEncoder):
            """JsonSchemaMixin encoder for Path fields"""

            # pylint: disable=too-few-public-methods
            @property
            def json_schema(self) -> dict:
                """Return the json schema"""
                return {"type": "path"}

        JsonSchemaMixin.register_field_encoders({Path: PathField()})

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            schema = cls.json_schema()

        sdefs = schema["definitions"]

        # pylint: disable=unused-variable
        sdefs["PrepRound"]["properties"]["prep_package_endpoints"] = {  # type: ignore # noqa: F841
            "package_endpoint_type": "prep",
            "type": "package_endpoint_array",
        }
        sdefs["PrepSettings"]["properties"]["prep_man_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "prep_man",
            "type": "package_endpoint_array",
        }
        sdefs["DedupeSettings"]["properties"]["dedupe_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "dedupe",
            "type": "package_endpoint_array",
        }
        sdefs["PrescreenSettings"]["properties"]["prescreen_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "prescreen",
            "type": "package_endpoint_array",
        }
        sdefs["PDFGetSettings"]["properties"]["pdf_get_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "pdf_get",
            "type": "package_endpoint_array",
        }
        sdefs["PDFGetSettings"]["properties"]["pdf_get_man_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "pdf_get_man",
            "type": "package_endpoint_array",
        }
        sdefs["PDFPrepSettings"]["properties"]["pdf_prep_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "pdf_prep",
            "type": "package_endpoint_array",
        }
        sdefs["PDFPrepSettings"]["properties"]["pdf_prep_man_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "pdf_prep_man",
            "type": "package_endpoint_array",
        }
        sdefs["ScreenSettings"]["properties"]["screen_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "screen",
            "type": "package_endpoint_array",
        }
        sdefs["DataSettings"]["properties"]["data_package_endpoints"] = {  # type: ignore
            "package_endpoint_type": "data",
            "type": "package_endpoint_array",
        }

        return schema


def __load_settings_from_dict(*, loaded_dict: dict) -> Settings:
    try:
        converters = {Path: Path, Enum: Enum}
        settings = from_dict(
            data_class=Settings,
            data=loaded_dict,
            config=dacite.Config(type_hooks=converters, cast=[Enum]),  # type: ignore
        )
        for source in settings.sources:
            if not str(source.filename).replace("\\", "/").startswith("data/search"):
                msg = f"Source filename does not start with data/search: {source.filename}"
                raise colrev_exceptions.InvalidSettingsError(msg=msg)

        filenames = [x.filename for x in settings.sources]
        if not len(filenames) == len(set(filenames)):
            non_unique = list({str(x) for x in filenames if filenames.count(x) > 1})
            msg = f"Non-unique source filename(s): {', '.join(non_unique)}"
            raise colrev_exceptions.InvalidSettingsError(msg=msg, fix_per_upgrade=False)

    except (ValueError, MissingValueError, WrongTypeError, AssertionError) as exc:
        raise colrev_exceptions.InvalidSettingsError(
            msg=str(exc)
        ) from exc  # pragma: no cover

    return settings


def load_settings(*, settings_path: Path) -> Settings:
    """Load the settings from file"""

    if not settings_path.is_file():
        raise colrev_exceptions.RepoSetupError()

    with open(settings_path, encoding="utf-8") as file:
        loaded_dict = json.load(file)

    return __load_settings_from_dict(loaded_dict=loaded_dict)


def save_settings(*, review_manager: colrev.review_manager.ReviewManager) -> None:
    """Save the settings"""

    exported_dict = asdict(
        review_manager.settings, dict_factory=colrev.env.utils.custom_asdict_factory
    )

    with open("settings.json", "w", encoding="utf-8") as outfile:
        json.dump(exported_dict, outfile, indent=4)
    review_manager.dataset.add_changes(path=Path("settings.json"))



#! /usr/bin/env python
"""Checker for html-tags."""
from __future__ import annotations

import re

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class HTMLTagChecker:
    """The HTMLTagChecker"""

    msg = "html-tags"
    __fields_to_check = [
        "title",
        "journal",
        "booktitle",
        "author",
        "publisher",
        "editor",
    ]

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the html-tags checks"""

        for key in self.__fields_to_check:
            if key in record.data:
                if re.search(r"&#\d+;", record.data[key]):
                    record.add_masterdata_provenance_note(key=key, note=self.msg)
                else:
                    record.remove_masterdata_provenance_note(key=key, note=self.msg)


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(HTMLTagChecker(quality_model))



#!/usr/bin/env python3
"""Advises users on the workflow (operations and collaboration)."""
from __future__ import annotations

import typing
from collections import Counter
from multiprocessing.dummy import Pool as ThreadPool
from pathlib import Path
from typing import Optional

import git
from git.exc import InvalidGitRepositoryError
from git.exc import NoSuchPathError

import colrev.record


class Advisor:
    """The CoLRev advisor guides users through the review process"""

    _next_step_description = {
        "retrieve": "Next step: retrieve metadata",
        "load": "Next step: Import search results",
        "prep": "Next step: Prepare records",
        "prep_man": "Next step: Prepare records (manually)",
        "dedupe": "Next step: Deduplicate records",
        "prescreen": "Next step: Prescreen records",
        "pdfs": "Next step: retrieve pdfs",
        "pdf_get": "Next step: Retrieve PDFs",
        "pdf_get_man": "Next step: Retrieve PDFs (manually)",
        "pdf_prep": "Next step: Prepare PDFs",
        "pdf_prep_man": "Next step: Prepare PDFs (manually)",
        "screen": "Next step: Screen records",
        "data": "Next step: Extract data/synthesize records",
    }

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        self.review_manager = review_manager

    def __append_merge_conflict_warning(
        self, *, collaboration_instructions: dict, git_repo: git.Repo
    ) -> None:
        found_a_conflict = False
        unmerged_blobs = git_repo.index.unmerged_blobs()
        for _, list_of_blobs in unmerged_blobs.items():
            for stage, _ in list_of_blobs:
                if stage != 0:
                    found_a_conflict = True
        if found_a_conflict:
            item = {
                "title": "Git merge conflict detected",
                "level": "WARNING",
                "msg": "To resolve:\n  1 https://docs.github.com/en/"
                + "pull-requests/collaborating-with-pull-requests/"
                + "addressing-merge-conflicts/resolving-a-merge-conflict-"
                + "using-the-command-line",
            }
            collaboration_instructions["items"].append(item)

    def __notify_non_staged_files(
        self, *, collaboration_instructions: dict, git_repo: git.Repo
    ) -> None:
        # Notify when changes in bib files are not staged
        # (this may raise unexpected errors)

        non_staged = [
            item.a_path
            for item in git_repo.index.diff(None)
            if item.a_path[-4:] == ".bib"
        ]
        if len(non_staged) > 0:
            item = {
                "title": f"Non-staged changes: {','.join(non_staged)}",
                "level": "WARNING",
            }
            collaboration_instructions["items"].append(item)

    def __add_sharing_notifications(
        self,
        *,
        collaboration_instructions: dict,
        status_stats: colrev.ops.status.StatusStats,
    ) -> None:
        # pylint: disable=too-many-branches

        share_stat_req = self.review_manager.settings.project.share_stat_req
        collaboration_instructions["SHARE_STAT_REQ"] = share_stat_req

        if self.review_manager.dataset.behind_remote():
            item = {
                "title": "Remote changes available on the server",
                "level": "WARNING",
                "msg": "Once you have committed your changes, get the latest "
                + "remote changes",
                "cmd_after": "git add FILENAME \n  git commit -m 'MSG' \n  "
                + "git pull --rebase",
            }
            collaboration_instructions["items"].append(item)

        if self.review_manager.dataset.remote_ahead():
            item = {
                "title": "Local changes not yet on the server",
                "level": "WARNING",
                "msg": "Once you have committed your changes, upload them "
                + "to the shared repository.",
                "cmd_after": "git push",
            }
            collaboration_instructions["items"].append(item)

        if share_stat_req == "NONE":
            collaboration_instructions["status"] = {
                "title": "Sharing: currently ready for sharing",
                "level": "SUCCESS",
                "msg": "",
                # If consistency checks pass -
                # if they didn't pass, the message wouldn't be displayed
            }

        if share_stat_req == "PROCESSED":
            if (
                0 == status_stats.currently.md_retrieved
                and 0 == status_stats.currently.md_imported
                and 0 == status_stats.currently.md_needs_manual_preparation
                and 0 == status_stats.currently.md_prepared
            ):
                collaboration_instructions["status"] = {
                    "title": "Sharing: currently ready for sharing",
                    "level": "SUCCESS",
                    "msg": "",
                    # If consistency checks pass -
                    # if they didn't pass, the message wouldn't be displayed
                }

            else:
                collaboration_instructions["status"] = {
                    "title": "Sharing: currently not ready for sharing",
                    "level": "WARNING",
                    "msg": "All records should be processed before sharing "
                    + "(see instructions above).",
                }

        # Note: if we use all(...) in the following,
        # we do not need to distinguish whether
        # a PRE_SCREEN or INCLUSION_SCREEN is needed
        # pylint: disable=too-many-boolean-expressions
        if share_stat_req == "SCREENED":
            if (
                0 == status_stats.currently.md_retrieved
                and 0 == status_stats.currently.md_imported
                and 0 == status_stats.currently.md_needs_manual_preparation
                and 0 == status_stats.currently.md_prepared
                and 0 == status_stats.currently.md_processed
                and 0 == status_stats.currently.rev_prescreen_included
                and 0 == status_stats.currently.pdf_needs_manual_retrieval
                and 0 == status_stats.currently.pdf_imported
                and 0 == status_stats.currently.pdf_needs_manual_preparation
                and 0 == status_stats.currently.pdf_prepared
            ):
                collaboration_instructions["status"] = {
                    "title": "Sharing: currently ready for sharing",
                    "level": "SUCCESS",
                    "msg": "",
                    # If consistency checks pass -
                    # if they didn't pass, the message wouldn't be displayed
                }

            else:
                collaboration_instructions["status"] = {
                    "title": "Sharing: currently not ready for sharing",
                    "level": "WARNING",
                    "msg": "All records should be screened before sharing "
                    + "(see instructions above).",
                }

        if share_stat_req == "COMPLETED":
            if 0 == status_stats.currently.non_completed:
                collaboration_instructions["status"] = {
                    "title": "Sharing: currently ready for sharing",
                    "level": "SUCCESS",
                    "msg": "",
                    # If consistency checks pass -
                    # if they didn't pass, the message wouldn't be displayed
                }
            else:
                collaboration_instructions["status"] = {
                    "title": "Sharing: currently not ready for sharing",
                    "level": "WARNING",
                    "msg": "All records should be completed before sharing "
                    + "(see instructions above).",
                }

    def __get_collaboration_instructions(
        self, *, status_stats: Optional[colrev.ops.status.StatusStats] = None
    ) -> dict:
        """Get instructions related to collaboration"""

        if status_stats is None:
            status_stats = self.review_manager.get_status_stats()

        collaboration_instructions: dict = {"items": []}
        git_repo = self.review_manager.dataset.get_repo()

        remote_connected = 0 != len(git_repo.remotes)
        if remote_connected:
            collaboration_instructions["title"] = "Versioning and collaboration"
        else:
            collaboration_instructions[
                "title"
            ] = "Versioning (not connected to shared repository)"
            item = {
                "title": "Project not yet shared",
                "level": "WARNING",
                "msg": "Please visit  https://github.com/new\n  "
                + "create an empty repository called  "
                + f"<USERNAME>/{self.review_manager.settings.project.title}\n  "
                + "and run git remote add origin  <REMOTE_URL>\n  git push origin main",
            }
            collaboration_instructions["items"].append(item)

        self.__append_merge_conflict_warning(
            collaboration_instructions=collaboration_instructions, git_repo=git_repo
        )
        if len(collaboration_instructions["items"]) > 0:
            # Don't append any other instructions.
            # Resolving the merge conflict is always prio 1
            return collaboration_instructions

        self.__notify_non_staged_files(
            collaboration_instructions=collaboration_instructions, git_repo=git_repo
        )

        if remote_connected:
            self.__add_sharing_notifications(
                collaboration_instructions=collaboration_instructions,
                status_stats=status_stats,
            )

        if 0 == len(collaboration_instructions["items"]):
            item = {
                "title": "Up-to-date",
                "level": "SUCCESS",
            }
            collaboration_instructions["items"].append(item)

        return collaboration_instructions

    def __append_initial_load_instruction(self, *, review_instructions: list) -> None:
        if not self.review_manager.dataset.records_file.is_file():
            instruction = {
                "msg": "To import, copy search results to the search directory.",
                "cmd": "colrev load",
            }
            if instruction["cmd"] not in [
                ri["cmd"] for ri in review_instructions if "cmd" in ri
            ]:
                review_instructions.append(instruction)

    def __append_operation_in_progress_instructions(
        self,
        *,
        review_instructions: list,
        status_stats: colrev.ops.status.StatusStats,
        current_origin_states_dict: dict,
    ) -> None:
        # If changes in RECORDS_FILE are staged, we need to detect the process type
        if self.review_manager.dataset.records_changed():
            # Detect and validate transitions
            transitioned_records = status_stats.get_transitioned_records(
                current_origin_states_dict=current_origin_states_dict
            )

            for transitioned_record in transitioned_records:
                if transitioned_record["dest"] == "no_source_state":
                    print(f"Error (no source_state): {transitioned_record}")
                    review_instructions.append(
                        {
                            "msg": "Resolve committed colrev_status "
                            + f"of {transitioned_record}",
                            "priority": "yes",
                        }
                    )
                if transitioned_record["operations_type"] == "invalid_transition":
                    msg = (
                        f"Resolve invalid transition ({transitioned_record['ID']}): "
                        + f"{transitioned_record['source']} to "
                        + f"{transitioned_record['dest']}"
                    )
                    if msg not in [ri["msg"] for ri in review_instructions]:
                        review_instructions.append(
                            {
                                "msg": msg,
                                "priority": "yes",
                            }
                        )

            in_progress_processes = status_stats.get_operation_in_progress(
                transitioned_records=transitioned_records
            )

            if len(in_progress_processes) == 1:
                instruction = {
                    "msg": f"Detected {in_progress_processes[0]} in progress. "
                    + "Complete this process",
                    "cmd": f"colrev {in_progress_processes[0]}",
                    "priority": "yes",
                }
                review_instructions.append(instruction)
            elif len(in_progress_processes) > 1:
                rec_str = ", ".join([x["origin"] for x in transitioned_records])
                instruction = {
                    "msg": "Detected multiple processes in progress "
                    + f"({', '.join(in_progress_processes)}). Complete one "
                    + "(save and revert the other) and commit before continuing!\n"
                    + f"  Records: {rec_str}",
                    # "cmd": f"colrev {in_progress_processes}",
                    "priority": "yes",
                }
                review_instructions.append(instruction)

    def __append_initial_operations(
        self, *, review_instructions: list, status_stats: colrev.ops.status.StatusStats
    ) -> bool:
        if not Path(self.review_manager.search_dir).iterdir():
            instruction = {
                "msg": "Add search results to data/search",
                "priority": "yes",
            }
            review_instructions.append(instruction)
            return True

        if status_stats.overall.md_retrieved == 0:
            instruction = {
                "msg": self._next_step_description["retrieve"],
                "cmd": "colrev retrieve",
                "priority": "yes",
            }
            review_instructions.append(instruction)
            return True

        if status_stats.currently.md_retrieved > 0:
            instruction = {
                "msg": self._next_step_description["retrieve"],
                "cmd": "colrev retrieve",
                "priority": "yes",
            }
            review_instructions.append(instruction)

            if not self.review_manager.verbose_mode:
                return True
        return False

    def __append_active_operations(
        self,
        *,
        status_stats: colrev.ops.status.StatusStats,
        current_origin_states_dict: dict,
        review_instructions: list,
    ) -> None:
        active_operations = status_stats.get_active_operations(
            current_origin_states_dict=current_origin_states_dict
        )

        priority_processing_operations = status_stats.get_priority_operations(
            current_origin_states_dict=current_origin_states_dict
        )

        for active_operation in active_operations:
            if active_operation in ["load", "prep", "dedupe"]:
                instruction = {
                    "msg": self._next_step_description["retrieve"],
                    "cmd": "colrev retrieve",
                }
            if active_operation in ["pdf_get", "pdf_prep"]:
                instruction = {
                    "msg": self._next_step_description["pdfs"],
                    "cmd": "colrev pdfs",
                }
            else:
                instruction = {
                    "msg": self._next_step_description[active_operation],
                    "cmd": f"colrev {active_operation.replace('_', '-')}",
                }
            if active_operation in priority_processing_operations:
                # keylist = [list(x.keys()) for x in review_instructions]
                # keys = [item for sublist in keylist for item in sublist]
                # if "priority" not in keys:
                instruction["priority"] = "yes"
            else:
                if (
                    self.review_manager.settings.project.delay_automated_processing
                    and not self.review_manager.verbose_mode
                ):
                    continue
            if instruction["cmd"] not in [
                ri["cmd"] for ri in review_instructions if "cmd" in ri
            ]:
                review_instructions.append(instruction)

    def __append_data_operation_advice(self, *, review_instructions: list) -> None:
        if (
            len(review_instructions) == 1
            or self.review_manager.verbose_mode
            or self.review_manager.settings.is_curated_masterdata_repo()
        ):
            if (
                "colrev data" in [ri["cmd"] for ri in review_instructions]
                or self.review_manager.settings.is_curated_masterdata_repo()
            ):
                for item in review_instructions.copy():
                    if item.get("cmd") == "colrev data":
                        review_instructions.remove(item)
                        break

                # review_instructions.pop(0)

                package_manager = self.review_manager.get_package_manager()
                check_operation = colrev.operation.CheckOperation(
                    review_manager=self.review_manager
                )
                for (
                    data_package_endpoint
                ) in self.review_manager.settings.data.data_package_endpoints:
                    endpoint_dict = package_manager.load_packages(
                        package_type=colrev.env.package_manager.PackageEndpointType.data,
                        selected_packages=[data_package_endpoint],
                        operation=check_operation,
                    )
                    endpoint = endpoint_dict[data_package_endpoint["endpoint"]]

                    advice = endpoint.get_advice(self.review_manager)  # type: ignore
                    if advice:
                        review_instructions.append(advice)

    def __append_next_operation_instructions(
        self,
        *,
        review_instructions: list,
        status_stats: colrev.ops.status.StatusStats,
        current_origin_states_dict: dict,
    ) -> None:
        if self.__append_initial_operations(
            review_instructions=review_instructions, status_stats=status_stats
        ):
            return

        self.__append_active_operations(
            status_stats=status_stats,
            current_origin_states_dict=current_origin_states_dict,
            review_instructions=review_instructions,
        )

        self.__append_data_operation_advice(review_instructions=review_instructions)

    def __get_missing_files(
        self, *, status_stats: colrev.ops.status.StatusStats
    ) -> list:
        # excluding pdf_not_available
        file_required_status = [
            colrev.record.RecordState.pdf_imported,
            colrev.record.RecordState.pdf_needs_manual_preparation,
            colrev.record.RecordState.pdf_prepared,
            colrev.record.RecordState.rev_excluded,
            colrev.record.RecordState.rev_included,
            colrev.record.RecordState.rev_synthesized,
        ]
        missing_files = []
        for record_dict in status_stats.records.values():
            if (
                record_dict["colrev_status"] in file_required_status
                and "file" not in record_dict
            ):
                missing_files.append(record_dict["ID"])

        return missing_files

    def __append_pdf_issue_instructions(
        self, *, status_stats: colrev.ops.status.StatusStats, review_instructions: list
    ) -> None:
        # Check pdf files
        if self.review_manager.settings.pdf_get.pdf_required_for_screen_and_synthesis:
            missing_files = self.__get_missing_files(status_stats=status_stats)
            if len(missing_files) > 0:
                review_instructions.append(
                    {
                        "msg": "record with colrev_status requiring a PDF file but missing "
                        + f"the path (file = ...): {missing_files}"
                    }
                )

            if len(missing_files) > 0:
                if len(missing_files) < 10:
                    non_existent_pdfs = ",".join(missing_files)
                else:
                    non_existent_pdfs = ",".join(missing_files[0:10] + ["..."])
                review_instructions.append(
                    {
                        "msg": f"record with broken file link ({non_existent_pdfs})."
                        " Use\n    colrev pdf-get --relink_pdfs"
                    }
                )

        pdfs_no_longer_available = []
        for record_dict in status_stats.records.values():
            if "file" in record_dict:
                if not (self.review_manager.path / Path(record_dict["file"])).is_file():
                    pdfs_no_longer_available.append(record_dict["file"])
        if pdfs_no_longer_available:
            review_instructions.append(
                {
                    "level": "WARNING",
                    "msg": f"PDF no longer available: {','.join(pdfs_no_longer_available)}",
                    "cmd": "colrev repare",
                }
            )

    def __append_iteration_completed_instructions(
        self, *, review_instructions: list, status_stats: colrev.ops.status.StatusStats
    ) -> None:
        if (
            status_stats.completeness_condition
            and status_stats.currently.md_retrieved > 0
        ):
            if not self.review_manager.dataset.has_untracked_search_records():
                instruction = {
                    "info": "Review iteration completed.",
                    "msg": "To start the next iteration of the review, "
                    + "add new search results (to data/search)",
                }
                review_instructions.append(instruction)
            else:
                instruction = {
                    "info": "Search results available for next iteration.",
                    "msg": "Next step: Import search results.",
                    "cmd": "colrev load",
                }
                review_instructions.append(instruction)

    def get_review_instructions(
        self, *, status_stats: Optional[colrev.ops.status.StatusStats] = None
    ) -> list:
        """Get instructions related to the review (operations)"""

        if status_stats is None:
            status_stats = self.review_manager.get_status_stats()

        review_instructions: typing.List[typing.Dict] = []
        current_origin_states_dict = self.review_manager.dataset.get_origin_state_dict()

        self.__append_initial_load_instruction(review_instructions=review_instructions)

        self.__append_operation_in_progress_instructions(
            review_instructions=review_instructions,
            status_stats=status_stats,
            current_origin_states_dict=current_origin_states_dict,
        )

        self.__append_next_operation_instructions(
            review_instructions=review_instructions,
            status_stats=status_stats,
            current_origin_states_dict=current_origin_states_dict,
        )

        self.__append_pdf_issue_instructions(
            status_stats=status_stats, review_instructions=review_instructions
        )

        self.__append_iteration_completed_instructions(
            review_instructions=review_instructions, status_stats=status_stats
        )

        return review_instructions

    # Note : no named arguments for multiprocessing
    def __append_registered_repo_instructions(self, registered_path: Path) -> dict:
        instruction = {}

        try:
            # Note : registered_path are other repositories (don't load from dataset.get_repo())
            git_repo = git.Repo(registered_path)

            # https://github.com/gitpython-developers/GitPython/issues/652#issuecomment-610511311
            origin = git_repo.remotes.origin

            if not origin.exists():
                raise AttributeError

            if git_repo.active_branch.tracking_branch() is None:
                raise AttributeError

            branch_name = str(git_repo.active_branch)
            tracking_branch_name = str(git_repo.active_branch.tracking_branch())
            # self.review_manager.logger.debug(f"{branch_name} - {tracking_branch_name}")

            behind_operation = branch_name + ".." + tracking_branch_name
            commits_behind = git_repo.iter_commits(behind_operation)
            nr_commits_behind = sum(1 for c in commits_behind)

            ahead_operation = tracking_branch_name + ".." + branch_name
            commits_ahead = git_repo.iter_commits(ahead_operation)
            nr_commits_ahead = sum(1 for c in commits_ahead)

            def pull_condition() -> bool:
                # behind_remote and not remote_ahead
                return nr_commits_behind > 0 and not nr_commits_ahead > 0

            def pull_rebase_condition() -> bool:
                # behind_remote and remote_ahead
                return nr_commits_behind > 0 and nr_commits_ahead > 0

            # Note: do not use named arguments (multiprocessing)
            if not Path(registered_path).is_dir():
                instruction = {
                    "msg": "Locally registered repo no longer exists.",
                    "cmd": f"colrev env --unregister {registered_path}",
                }

            elif "curated_metadata" in str(registered_path):
                if pull_condition():
                    instruction = {
                        "msg": "Updates available for curated repo "
                        f"({registered_path}).",
                        "cmd": "colrev env --pull",
                    }

                elif pull_rebase_condition():
                    instruction = {
                        "msg": "Local/remote branch diverged for curated repo "
                        f"({registered_path}).",
                        "cmd": f"cd '{registered_path}' && git pull --rebase",
                    }

        except (AttributeError, NoSuchPathError, InvalidGitRepositoryError):
            pass
        return instruction

    def __append_download_outlets_instruction(
        self,
        environment_manager: colrev.env.environment_manager.EnvironmentManager,
        environment_instructions: list,
    ) -> None:
        """Get instructions related to downloading outlets (resources)"""

        # pylint: disable=too-many-locals

        with open(self.review_manager.dataset.records_file, encoding="utf8") as file:
            outlets = []
            for line in file.readlines():
                if line.lstrip()[:8] == "journal ":
                    journal = line[line.find("{") + 1 : line.rfind("}")]
                    outlets.append(journal)
                if line.lstrip()[:10] == "booktitle ":
                    booktitle = line[line.find("{") + 1 : line.rfind("}")]
                    outlets.append(booktitle)

        outlet_counter: typing.List[typing.Tuple[str, int]] = [
            (j, x) for j, x in Counter(outlets).most_common(10) if x > 5
        ]
        selected = []
        cumulative = 0.0
        for candidate, freq in outlet_counter:
            selected.append((candidate, freq))
            cumulative += freq / len(outlets)
            if cumulative > 0.7:
                break
        if len(selected) > 0:
            curated_outlets = environment_manager.get_curated_outlets()
            selected_journals = [
                (candidate, freq)
                for candidate, freq in selected
                if candidate not in curated_outlets + ["", "UNKNOWN"]
            ]

            journals = "\n   - " + "\n   - ".join(
                [
                    f"{candidate} ({round((freq/len(outlets))*100, 2)}%)"
                    for candidate, freq in selected_journals
                ]
            )

            if len(selected_journals) > 0:
                instruction = {
                    "msg": "Search and download curated metadata for your "
                    "project (if available). \n  The most common journals in "
                    f"your project are {journals}.\n"
                    "  They may be available at "
                    "https://github.com/CoLRev-curations",
                }
                environment_instructions.append(instruction)

    def __get_environment_instructions(
        self, *, status_stats: colrev.ops.status.StatusStats
    ) -> list:
        """Get instructions related to the CoLRev environment"""

        environment_manager = self.review_manager.get_environment_manager()

        environment_instructions: list[dict] = []

        if status_stats.currently.md_imported > 10:
            self.__append_download_outlets_instruction(
                environment_manager=environment_manager,
                environment_instructions=environment_instructions,
            )

        local_repos = environment_manager.local_repos()
        registered_paths = [Path(x["repo_source_path"]) for x in local_repos]
        # Note : we can use many parallel processes
        # because __append_registered_repo_instructions mainly waits for the network
        # it does not use a lot of CPU capacity
        pool = ThreadPool(50)
        add_instructions = pool.map(
            self.__append_registered_repo_instructions, registered_paths
        )

        environment_instructions += list(filter(None, add_instructions))

        if len(list(self.review_manager.corrections_path.glob("*.json"))) > 0:
            instruction = {
                "msg": "Corrections to share with curated repositories.",
                "cmd": "colrev push -r",
            }
            environment_instructions.append(instruction)

        return environment_instructions

    def get_instructions(
        self, *, status_stats: Optional[colrev.ops.status.StatusStats] = None
    ) -> dict:
        """Get all instructions on the project"""

        if status_stats is None:
            status_stats = self.review_manager.get_status_stats()

        instructions = {
            "review_instructions": self.get_review_instructions(
                status_stats=status_stats
            ),
            "environment_instructions": self.__get_environment_instructions(
                status_stats=status_stats
            ),
            "collaboration_instructions": self.__get_collaboration_instructions(
                status_stats=status_stats
            ),
        }
        # self.review_manager.logger.debug(
        #     f"instructions: {self.review_manager.p_printer.pformat(instructions)}"
        # )
        if self.review_manager.shell_mode:
            for category in instructions.values():
                for item in category:
                    if "cmd" in item:
                        item["cmd"] = item["cmd"].replace("colrev ", "")
        return instructions

    def get_sharing_instructions(self) -> dict:
        """Get instructions related to sharing the project"""

        collaboration_instructions = self.__get_collaboration_instructions()

        status_code = not all(
            x["level"] in ["SUCCESS", "WARNING"]
            for x in collaboration_instructions["items"]
        )

        msgs = "\n ".join(
            [
                x["level"] + x["title"] + x.get("msg", "")
                for x in collaboration_instructions["items"]
            ]
        )
        return {"msg": msgs, "status": status_code}



#!/usr/bin/env python3
"""Scripts for the load process."""
from __future__ import annotations

import typing
from pathlib import Path

import colrev.record
import colrev.ui_cli.cli_colors as colors


def __select_source(
    *,
    source_candidates: list,
) -> dict:
    print(f"{colors.ORANGE}Select search source{colors.END}:")
    for i, heuristic_source in enumerate(source_candidates):
        highlight_color = ""
        if heuristic_source["confidence"] >= 0.7:
            highlight_color = colors.GREEN
        elif heuristic_source["confidence"] >= 0.5:
            highlight_color = colors.ORANGE
        print(
            f"{highlight_color}{i+1} "
            f"(confidence: {round(heuristic_source['confidence'], 2)}):"
            f" {heuristic_source['source_candidate'].endpoint}{colors.END}"
        )

    while True:
        selection = input("select nr")
        if not selection.isdigit():
            continue
        if int(selection) in range(1, len(source_candidates) + 1):
            heuristic_source = source_candidates[int(selection) - 1]
            return heuristic_source


def __select_source_from_heuristics(
    *, filename: Path, source_candidates: list
) -> colrev.settings.SearchSource:
    print(f"Search file {filename}")
    if 1 == len(source_candidates):
        heuristic_source = source_candidates[0]
    else:
        heuristic_source = __select_source(source_candidates=source_candidates)

    if "colrev.unknown_source" == heuristic_source["source_candidate"].endpoint:
        cmd = "Enter the search query (or NA)".ljust(25, " ") + ": "
        query_input = ""
        query_input = input(cmd)
        if query_input not in ["", "NA"]:
            heuristic_source["source_candidate"].search_parameters = {
                "query": query_input
            }
        else:
            heuristic_source["source_candidate"].search_parameters = {}

    print(f"Source name: {heuristic_source['source_candidate'].endpoint}")

    heuristic_source["source_candidate"].comment = None

    return heuristic_source["source_candidate"]


def select_new_source(
    *, heuristic_result_list: dict
) -> typing.List[colrev.settings.SearchSource]:
    """Select the new source from the heuristic_result_list."""

    new_sources = []
    for filename, source_candidates in heuristic_result_list.items():
        new_sources.append(
            __select_source_from_heuristics(
                filename=filename, source_candidates=source_candidates
            )
        )
    return new_sources



#! /usr/bin/env python
"""Creation of screenshots (PDFs) for online ENTRYTYPES"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.record

# pylint: disable=duplicate-code

if TYPE_CHECKING:
    import colrev.ops.pdf_get

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PDFGetPackageEndpointInterface)
@dataclass
class WebsiteScreenshot(JsonSchemaMixin):
    """Get PDFs from website screenshot (for "online" ENTRYTYPES)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def get_pdf(
        self, pdf_get_operation: colrev.ops.pdf_get.PDFGet, record: colrev.record.Record
    ) -> colrev.record.Record:
        """Get a PDF of the website (screenshot)"""

        if record.data["ENTRYTYPE"] != "online":
            return record

        screenshot_service = pdf_get_operation.review_manager.get_screenshot_service()
        screenshot_service.start_screenshot_service()

        pdf_filepath = pdf_get_operation.review_manager.PDF_DIR_RELATIVE / Path(
            f"{record.data['ID']}.pdf"
        )
        record = screenshot_service.add_screenshot(
            record=record, pdf_filepath=pdf_filepath
        )

        if "file" in record.data:
            pdf_get_operation.import_pdf(record=record)

        return record



#! /usr/bin/env python
"""CoLRev data operation: extract data, analyze, and synthesize."""
from __future__ import annotations

from pathlib import Path
from typing import Optional

import pandas as pd

import colrev.operation
import colrev.ops.built_in.pdf_prep.tei_prep
import colrev.record
import colrev.ui_cli.cli_colors as colors


class Data(colrev.operation.Operation):
    """Class supporting structured and unstructured
    data extraction, analysis and synthesis"""

    __pad = 0

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.data,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.package_manager = self.review_manager.get_package_manager()

    def get_record_ids_for_synthesis(self, records: dict) -> list:
        """Get the IDs of records for the synthesis"""
        return [
            ID
            for ID, record in records.items()
            if record["colrev_status"]
            in [
                colrev.record.RecordState.rev_included,
                colrev.record.RecordState.rev_synthesized,
            ]
        ]

    def reading_heuristics(self) -> list:
        """Determine heuristics for the reading process"""

        enlit_list = []
        records = self.review_manager.dataset.load_records_dict()
        for relevant_record_id in self.get_record_ids_for_synthesis(records):
            enlit_status = str(records[relevant_record_id]["colrev_status"])
            enlit_status = enlit_status.replace("rev_included", "").replace(
                "rev_synthesized", "synthesized"
            )
            enlit_list.append(
                {
                    "ID": relevant_record_id,
                    "score": 0,
                    "score_intensity": 0,
                    "colrev_status": enlit_status,
                }
            )

        tei_path = colrev.ops.built_in.pdf_prep.tei_prep.TEIPDFPrep.TEI_PATH_RELATIVE
        required_records_ids = self.get_record_ids_for_synthesis(records)

        missing = []
        for required_records_id in required_records_ids:
            tei_file = tei_path / Path(f"{required_records_id}.tei.xml")
            if not tei_file.is_file():
                missing.append(required_records_id)

            tei_doc = self.review_manager.get_tei(
                tei_path=tei_file,
            )
            tei_doc.mark_references(records=records)
            data = tei_doc.get_tei_str()
            for enlit_item in enlit_list:
                id_string = f'ID="{enlit_item["ID"]}"'
                if id_string in data:
                    enlit_item["score"] += 1
                enlit_item["score_intensity"] += data.count(id_string)

        if len(missing) > 0:
            print(f"Records with missing tei file: {missing}")

        enlit_list = sorted(enlit_list, key=lambda d: d["score"], reverse=True)

        return enlit_list

    def profile(self) -> None:
        """Create a profile of the sample"""

        self.review_manager.logger.info("Create sample profile")

        def prep_records(*, records: dict) -> pd.DataFrame:
            for record in records.values():
                record["outlet"] = record.get("journal", record.get("booktitle", "NA"))

            records_df = pd.DataFrame.from_records(list(records.values()))

            required_cols = [
                "ID",
                "ENTRYTYPE",
                "author",
                "title",
                "journal",
                "booktitle",
                "outlet",
                "year",
                "volume",
                "number",
                "pages",
                "doi",
            ]
            available_cols = records_df.columns.intersection(list(set(required_cols)))
            cols = [x for x in required_cols if x in available_cols]
            records_df = records_df[cols]
            return records_df

        def prep_observations(
            *, prepared_records_df: pd.DataFrame, records: dict
        ) -> pd.DataFrame:
            included_papers = [
                ID
                for ID, record in records.items()
                if record["colrev_status"]
                in [
                    colrev.record.RecordState.rev_synthesized,
                    colrev.record.RecordState.rev_included,
                ]
                and record.get("year", "UNKNOWN").isdigit()
            ]
            observations = prepared_records_df[
                prepared_records_df["ID"].isin(included_papers)
            ].copy()
            observations.year = observations.year.astype(int)
            missing_outlet = observations[observations["outlet"].isnull()][
                "ID"
            ].tolist()
            if len(missing_outlet) > 0:
                self.review_manager.logger.info(f"No outlet: {missing_outlet}")
            return observations

        # if not status.get_completeness_condition():
        #     self.review_manager.logger.warning(
        #  f"{colors.RED}Sample not completely processed!{colors.END}")

        records = self.review_manager.dataset.load_records_dict()

        output_dir = self.review_manager.path / Path("output")
        output_dir.mkdir(exist_ok=True)

        prepared_records_df = prep_records(records=records)  # .values()
        observations = prep_observations(
            prepared_records_df=prepared_records_df, records=records
        )

        if observations.empty:
            self.review_manager.logger.info("No sample/observations available")
            return

        self.review_manager.logger.info("Generate output/sample.csv")
        observations.to_csv(output_dir / Path("sample.csv"), index=False)

        tabulated = pd.pivot_table(
            observations[["outlet", "year"]],
            index=["outlet"],
            columns=["year"],
            aggfunc=len,
            fill_value=0,
            margins=True,
        )
        # Fill missing years with 0 columns
        years = range(
            min(e for e in tabulated.columns if isinstance(e, int)),
            max(e for e in tabulated.columns if isinstance(e, int)) + 1,
        )
        for year in years:
            if year not in tabulated.columns:
                tabulated[year] = 0
        year_list = list(years)
        year_list.extend(["All"])  # type: ignore
        tabulated = tabulated[year_list]
        tabulated.sort_values(by=("All"), ascending=True, inplace=True)

        self.review_manager.logger.info("Generate profile output/journals_years.csv")
        tabulated.to_csv(output_dir / Path("journals_years.csv"))

        tabulated = pd.pivot_table(
            observations[["ENTRYTYPE", "year"]],
            index=["ENTRYTYPE"],
            columns=["year"],
            aggfunc=len,
            fill_value=0,
            margins=True,
        )
        self.review_manager.logger.info("Generate output/ENTRYTYPES.csv")
        tabulated.to_csv(output_dir / Path("ENTRYTYPES.csv"))

        self.review_manager.logger.info(f"Files are available in {output_dir.name}")

    def setup_custom_script(self) -> None:
        """Setup a custom data script"""

        filedata = colrev.env.utils.get_package_file_content(
            file_path=Path("template/custom_scripts/custom_data_script.py")
        )
        if filedata:
            with open("custom_data_script.py", "w", encoding="utf-8") as file:
                file.write(filedata.decode("utf-8"))

        self.review_manager.dataset.add_changes(path=Path("custom_data_script.py"))

        new_data_endpoint = {"endpoint": "custom_data_script"}

        self.review_manager.settings.data.data_package_endpoints.append(
            new_data_endpoint
        )
        self.review_manager.save_settings()

    def __pre_data(self, *, records: dict, silent_mode: bool) -> None:
        if not silent_mode:
            self.review_manager.logger.info("Data")
            self.review_manager.logger.info(
                "The data operation covers different forms of data extraction, "
                "analysis, and synthesis."
            )
            self.review_manager.logger.info(
                "See https://colrev.readthedocs.io/en/latest/manual/data/data.html"
            )
        self.__pad = min((max(len(ID) for ID in list(records.keys()) + [""]) + 2), 35)

    def __get_synthesized_record_status_matrix(self, *, records: dict) -> dict:
        included = self.get_record_ids_for_synthesis(records)

        # TBD: do we assume that records are not changed by the processes?
        # records = self.review_manager.dataset.load_records_dict()

        # synthesized_record_status_matrix (paper IDs x endpoint):
        # each endpoint sets synthesized = True/False
        # and if a paper has synthesized=True in all fields,
        # its overall status is set to synthesized
        # Some endpoints may always set synthesized
        default_row = {
            df["endpoint"]: False
            for df in self.review_manager.settings.data.data_package_endpoints
        }
        synthesized_record_status_matrix = {ID: default_row.copy() for ID in included}

        # if self.review_manager.verbose_mode:
        #     self.review_manager.p_printer.pprint(synthesized_record_status_matrix)
        return synthesized_record_status_matrix

    def __update_record_status_matrix(
        self, *, records: dict, synthesized_record_status_matrix: dict
    ) -> bool:
        records_changed = False
        for (
            record_id,
            individual_status_dict,
        ) in synthesized_record_status_matrix.items():
            if all(x for x in individual_status_dict.values()):
                if (
                    records[record_id]["colrev_status"]
                    != colrev.record.RecordState.rev_synthesized
                ):
                    if self.review_manager.verbose_mode:
                        self.review_manager.report_logger.info(
                            f" {record_id}".ljust(self.__pad, " ")
                            + "set colrev_status to synthesized"
                        )
                        self.review_manager.logger.info(
                            f" {record_id}".ljust(self.__pad, " ")
                            + "set colrev_status to synthesized"
                        )

                if (
                    colrev.record.RecordState.rev_synthesized
                    != records[record_id]["colrev_status"]
                ):
                    records[record_id].update(
                        colrev_status=colrev.record.RecordState.rev_synthesized
                    )
                    records_changed = True
            else:
                if (
                    colrev.record.RecordState.rev_included
                    != records[record_id]["colrev_status"]
                ):
                    records[record_id].update(
                        colrev_status=colrev.record.RecordState.rev_included
                    )
                    records_changed = True
        return records_changed

    def __post_data(self, *, silent_mode: bool) -> None:
        # if self.review_manager.verbose_mode:
        #     self.review_manager.p_printer.pprint(synthesized_record_status_matrix)

        if self.review_manager.verbose_mode and not silent_mode:
            print()

        if not silent_mode:
            self.review_manager.logger.info(
                f"{colors.GREEN}Completed data operation{colors.END}"
            )
        if self.review_manager.in_ci_environment():
            print("\n\n")

    @colrev.operation.Operation.decorate()
    def main(
        self,
        *,
        selection_list: Optional[list] = None,
        records: Optional[dict] = None,
        silent_mode: bool = False,
    ) -> dict:
        """Data operation (main entrypoint)

        silent_mode: for review_manager checks
        """

        if not records:
            records = self.review_manager.dataset.load_records_dict()

        self.__pre_data(records=records, silent_mode=silent_mode)

        synthesized_record_status_matrix = self.__get_synthesized_record_status_matrix(
            records=records
        )

        for (
            data_package_endpoint
        ) in self.review_manager.settings.data.data_package_endpoints:
            if selection_list:
                if not any(
                    x in data_package_endpoint["endpoint"] for x in selection_list
                ):
                    continue

            if not silent_mode:
                print()
                self.review_manager.logger.info(
                    f"Data: {data_package_endpoint['endpoint'].replace('colrev.', '')}"
                )

            endpoint_dict = self.package_manager.load_packages(
                package_type=colrev.env.package_manager.PackageEndpointType.data,
                selected_packages=[data_package_endpoint],
                operation=self,
                only_ci_supported=self.review_manager.in_ci_environment(),
            )

            if data_package_endpoint["endpoint"] not in endpoint_dict:
                self.review_manager.logger.info(
                    f'Skip {data_package_endpoint["endpoint"]} (not available)'
                )
                continue

            endpoint = endpoint_dict[data_package_endpoint["endpoint"]]

            endpoint.update_data(  # type: ignore
                self, records, synthesized_record_status_matrix, silent_mode=silent_mode
            )

            endpoint.update_record_status_matrix(  # type: ignore
                self,
                synthesized_record_status_matrix,
                data_package_endpoint["endpoint"],
            )

            if self.review_manager.verbose_mode and not silent_mode:
                msg = f"Updated {endpoint.settings.endpoint}"  # type: ignore
                self.review_manager.logger.info(msg)

        records_status_changed = self.__update_record_status_matrix(
            records=records,
            synthesized_record_status_matrix=synthesized_record_status_matrix,
        )
        if records_status_changed:
            self.review_manager.dataset.save_records_dict(records=records)
            self.review_manager.dataset.add_record_changes()

        self.__post_data(silent_mode=silent_mode)

        no_endpoints_registered = 0 == len(
            self.review_manager.settings.data.data_package_endpoints
        )

        return {
            "ask_to_commit": self.review_manager.dataset.has_changes(),
            "no_endpoints_registered": no_endpoints_registered,
        }
